# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import traceback
import logging


def import_class(import_str):
    """Returns a class from a string including module and class"""
    mod_str, _, class_str = import_str.rpartition('.')
    logging.debug("Trying to import module %s", mod_str)
    try:
        __import__(mod_str)
        logging.debug("Trying to get class %s", class_str)
        return getattr(sys.modules[mod_str], class_str)
    except AttributeError:
        raise ImportError("No '%s' class in %s" % (class_str, mod_str))
    except ValueError as e:
        logging.debug("Exception occurred during import: %s", e)
        raise ImportError('Class %s cannot be found (%s)' %
                          (class_str,
                           traceback.format_exception(*sys.exc_info())))
    except Exception as e:
        logging.debug(
            "Exception while trying to import class %s: %s", import_str, e)
        raise


def import_object(import_str, *args, **kwargs):
    """Import a class and return an instance of it."""
    return import_class(import_str)(*args, **kwargs)

# Method for synchronization


def synchronized(method):

    def synchronized_method(self, *arg, **kws):
        with self.lock:
            return method(self, *arg, **kws)

    return synchronized_method


def enum(*sequential, **named):
    enums = dict(zip(sequential, range(len(sequential))), **named)
    reverse = dict((value, key) for key, value in enums.iteritems())
    enums['reverse'] = reverse
    return type('Enum', (), enums)


def getBoolean(string):
    '''
    return True is string represents boolean true ("true","yes","on","1"),
    False if not
    '''
    if isinstance(string, bool):
        return string
    assert(isinstance(string, str))
    return (string.lower() in ["true", "yes", "on", "1"])


def plural(x):
    if len(x) > 1:
        return "s"
    else:
        return ""



# encoding: utf-8
"""
community.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
Modified by Orange - 2014
"""

from struct import pack,unpack

import socket

from bagpipe.exabgp.message.update.attribute import AttributeID,Flag,Attribute

from bagpipe.exabgp.structure.asn import ASN 


# =================================================================== Community

class Community (object):
	def __init__ (self,community):
		self.community = community
		if community == 0xFFFFFF01:
			self._str = 'no-export'
		elif community == 0xFFFFFF02:
			self._str = 'no-advertise'
		elif community == 0xFFFFFF03:
			self._str = 'no-export-subconfed'
		else:
			self._str = "%d:%d" % (community >> 16, community & 0xFFFF)

	def pack (self):
		return pack('!L',self.community)

	def __str__ (self):
		return self._str

	def __len__ (self):
		return 4

	def __cmp__ (self,other):
		if ( isinstance(other,Community) and
			self.community == other.community):
			return 0
		else:
			return -1

# =================================================================== Communities (8)

class Communities (Attribute):
	ID = AttributeID.COMMUNITY
	FLAG = Flag.TRANSITIVE|Flag.OPTIONAL
	MULTIPLE = False

	def __init__ (self,communities=None):
		# Must be None as = param is only evaluated once
		if communities:
			self.communities = communities
		else:
			self.communities = []

	def add(self,data):
		return self.communities.append(data)

	def pack (self):
		if len(self.communities):
			return self._attribute(''.join([c.pack() for c in self.communities]))
		return ''

	def __str__ (self):
		l = len(self.communities)
		if l > 1:
			return "[ %s ]" % " ".join(str(community) for community in self.communities)
		if l == 1:
			return str(self.communities[0])
		return ""

	def __repr__ (self):
		return str(self)

# =================================================================== ECommunity

#def to_ECommunity (data):
#	separator = data.find(':')
#	if separator > 0:
#		# XXX: Check that the value do not overflow 16 bits
#		return ECommunity((int(data[:separator])<<16) + int(data[separator+1:]))
#	elif len(data) >=2 and data[1] in 'xX':
#		return ECommunity(long(data,16))
#	else:
#		return ECommunity(long(data))


# http://www.iana.org/assignments/bgp-extended-communities


# MUST ONLY raise ValueError
def to_ExtendedCommunity (data):
	command,ga,la = data.split(':')

	if command == 'origin':
		subtype = chr(0x03)
	elif command == 'target':
		subtype = chr(0x02)
	else:
		raise ValueError('invalid extended community %s (only origin or target are supported) ' % command)

	gc = ga.count('.')
	lc = la.count('.')
	if gc == 0 and lc == 3:
		# ASN first, IP second
		header = chr(0x40)
		global_admin = pack('!H',int(ga))
		local_admin = pack('!BBBB',*[int(_) for _ in la.split('.')])
	elif gc == 3 and lc == 0:
		# IP first, ASN second
		header = chr(0x41)
		global_admin = pack('!BBBB',*[int(_) for _ in ga.split('.')])
		local_admin = pack('!H',int(la))
	else:
		raise ValueError('invalid extended community %s ' % data)

	return ECommunity(header+subtype+global_admin+local_admin)

class ECommunity (object):
	ID = AttributeID.EXTENDED_COMMUNITY
	FLAG = Flag.TRANSITIVE|Flag.OPTIONAL
	MULTIPLE = False

	# size of value for data (boolean: is extended)
	length_value = {False:7, True:6}
	name = {False: 'regular', True: 'extended'}

	def __init__ (self,community):
		# Two top bits are iana and transitive bits
		self.community = community

	def iana (self):
		return not not (self.community[0] & 0x80)

	def transitive (self):
		return not not (self.community[0] & 0x40)

	def pack (self):
		return self.community

	def __str__ (self):
		# 30/02/12 Quagga communities for soo and rt are not transitive when 4360 says they must be, hence the & 0x0F
		community_type = ord(self.community[0]) & 0x0F
		community_stype = ord(self.community[1])
		# Target
		if community_stype == 0x02:
			return repr( RouteTarget.unpackFrom( self.community ) )
#			if community_type in (0x00,0x02):
#				asn = unpack('!H',self.community[2:4])[0]
#				ip = '%s.%s.%s.%s' % unpack('!BBBB',self.community[4:])
#				return "target:%d:%s" % (asn,ip)
#			if community_type == 0x01:
#				ip = '%s.%s.%s.%s' % unpack('!BBBB',self.community[2:6])
#				asn = unpack('!H',self.community[6:])[0]
#				return "target:%s:%d" % (ip,asn)
		# Origin
		if community_stype == 0x03:
			if community_type in (0x00,0x02):
				asn = unpack('!H',self.community[2:4])[0]
				ip = unpack('!L',self.community[4:])[0]
				return "origin:%d:%s" % (asn,ip)
			if community_type == 0x01:
				ip = '%s.%s.%s.%s' % unpack('!BBBB',self.community[2:6])
				asn = unpack('!H',self.community[6:])[0]
				return "origin:%s:%d" % (ip,asn)
			
		# Encapsulation 
		if community_stype == 0x0c:
			return repr( Encapsulation.unpackFrom(self.community))
				
		h = 0x00
		for byte in self.community:
			h <<= 8
			h += ord(byte)
		return "0x%016X" % h

	def __repr__ (self):
		return str(self)

	def __len__ (self):
		return 8

	def __cmp__ (self,other):
		return cmp(self.community,other.community)

	@staticmethod
	def unpackFrom(data):
		community_stype = ord(data[1])
		if community_stype == 0x02:
			return RouteTarget.unpackFrom( data )
		elif community_stype == 0x0c:
			return Encapsulation.unpackFrom( data )
		else:
			return ECommunity(data)
		

class RouteTarget(ECommunity):

	def __init__(self,asn,ip,number):
		assert (asn is None or ip is None)
		assert (asn is not None or ip is not None)
		
		if not asn is None:
			self.asn = asn
			self.number = number
			self.ip = ""
		else:
			self.ip = ip
			self.number = number
			self.asn = 0
			
		self.community = self.pack()
	
	def pack(self):
		if self.asn is not None:
			# type could also be 0x02 -> FIXME check RFC
			#return pack( 'BB!H!L', 0x00,0x02, self.asn, self.number)
			return pack( '!BBHL', 0x00,0x02, self.asn, self.number)
		else:
			encoded_ip = socket.inet_pton( socket.AF_INET, self.ip )            
			return pack( '!BB4sH', 0x01,0x02, encoded_ip, self.number)

	def __str__(self):
		if self.asn is not None:
			return "target:%s:%d" % ( str(self.asn), self.number  )
		else:
			return "target:%s:%d" % ( self.ip, self.number )

	def __cmp__(self,other):
		if ( isinstance(other,RouteTarget) and
			self.community == other.community ):
			return 0
		else:
			return -1
		
	def __hash__(self):
		return hash(self.community)

	@staticmethod
	def unpackFrom(data):
		type_  = ord(data[0]) & 0x0F
		stype = ord(data[1])
		data = data[2:]
	
		if stype == 0x02:  #FIXME, unclean
			if type_ in (0x00,0x02):
				asn,number    = unpack('!HL', data[:6] )
				return RouteTarget( ASN(asn) ,None,number)
			if type_ == 0x01:
				ip = socket.inet_ntop( data[0:4] )
				number = unpack('!H',data[4:6])[0]
				return RouteTarget(None,ip,number)


		



# RFC 5512, section 4.5

class Encapsulation(ECommunity): 

	ECommunity_TYPE = 0x03
	ECommunity_SUBTYPE = 0x0c

	DEFAULT=0
	# https://www.iana.org/assignments/bgp-parameters/bgp-parameters.xhtml#tunnel-types
	L2TPv3=1
	GRE=2
	IPIP=7
	VXLAN=8
	NVGRE=9
	MPLS=10
	VXLAN_GPE=12
	MPLS_UDP=13
	
	encapType2String={
					L2TPv3: "L2TPv3",
					GRE:    "GRE",
					VXLAN:  "VXLAN",
					NVGRE:  "NVGRE",
					IPIP:   "IP-in-IP",
					MPLS:	"MPLS",
					MPLS_UDP: "MPLS-in-UDP",
					VXLAN_GPE: "VXLAN-GPE",
					DEFAULT:"Default"
				}

	def __init__(self,tunnel_type):
		assert(isinstance(tunnel_type,int))
		self.tunnel_type = tunnel_type
		self.community = self.pack()

	def __str__(self):
		if self.tunnel_type in Encapsulation.encapType2String:
			return "Encap:" + Encapsulation.encapType2String[self.tunnel_type]
		else:
			return "Encap:(unknown:%d)" % self.tunnel_type
	
	def __hash__(self):
		return hash(self.community)

	def __cmp__(self,other):
		if isinstance(other,Encapsulation):
			return cmp(self.tunnel_type,other.tunnel_type)
		
	def pack(self):
		return pack("!BBHHH",
				Encapsulation.ECommunity_TYPE,
				Encapsulation.ECommunity_SUBTYPE,
				0,
				0,
				self.tunnel_type)
	
	@staticmethod
	def unpackFrom(data):
		type_  = ord(data[0]) & 0x0F
		stype = ord(data[1])
		data = data[2:]
	
		assert(type_==Encapsulation.ECommunity_TYPE)
		assert(stype==Encapsulation.ECommunity_SUBTYPE)
		assert(len(data)==6)
	
		tunnel_type=unpack('!H',data[4:6])[0]
		
		return Encapsulation(tunnel_type)


# =================================================================== ECommunities (16)

#def new_ECommunities (data):
#	communities = ECommunities()
#	while data:
#		ECommunity = unpack(data[:8])
#		data = data[8:]
#		communities.add(ECommunity(ECommunity))
#	return communities

class ECommunities (Communities):
	ID = AttributeID.EXTENDED_COMMUNITY

# =================================================================== FlowSpec Defined Extended Communities

def _to_FlowCommunity (action,data):
	return ECommunity(pack('!H',action) + data[:6])

# rate is bytes/seconds
def to_FlowTrafficRate (asn,rate):
	return _to_FlowCommunity (0x8006,pack('!H',asn)[:2]+pack('!f',rate))

def to_RouteOriginCommunity (asn,number,hightype=0x01):
	return ECommunity(chr(hightype) + chr(0x03) + pack('!H',asn) + pack('!L',number))

# VRF is ASN:Long
def to_RouteTargetCommunity_00 (asn,number):
	return ECommunity(chr(0x00) + chr(0x02) + pack('!H',asn) + pack('!L',number))

# VRF is A.B.C.D:Short
def to_RouteTargetCommunity_01 (ipn,number):
	return ECommunity(chr(0x01) + chr(0x02) + pack('!L',ipn) + pack('!H',number))

#def to_FlowAction (sample,terminal):
#	bitmask = chr(0)
#	if terminal: bitmask += 0x01
#	if sample: bitmask += 0x02
#	return _to_FlowCommunity (0x8007,chr(0)*5+bitmask)
#
## take a string representing a 6 bytes long hexacedimal number like "0x123456789ABC"
#def to_FlowRedirect (bitmask):
#	route_target = ''
#	for p in range(2,14,2): # 2,4,6,8,10,12
#		route_target += chr(int(bitmask[p:p+2],16))
#	return _to_FlowCommunity (0x8008,route_target)
#
#def to_FlowMark (dscp):
#	return _to_FlowCommunity (0x8009,chr(0)*5 + chr(dscp))
#
#def to_ASCommunity (subtype,asn,data,transitive):
#	r = chr(0x00)
#	if transitive: r += chr(0x40)
#	return ECommunity(r + chr(subtype) + pack('!H',asn) + ''.join([chr(c) for c in data[:4]]))
#
#import socket
#def to_IPv4Community (subtype,data,transitive):
#	r = chr(0x01)
#	if transitive: r += chr(0x40)
#	return ECommunity(r + chr(subtype) + socket.inet_pton(socket.AF_INET,ipv4) + ''.join([chr(c) for c in data[:2]]))
#
#def to_OpaqueCommunity (subtype,data,transitive):
#	r = chr(0x03)
#	if transitive: r += chr(0x40)
#	return ECommunity(r + chr(subtype) + ''.join([chr(c) for c in data[:6]]))

# See RFC4360
# 0x00, 0x02 Number is administrated by a global authority
# Format is asn:route_target (2 bytes:4 bytes)
# 0x01, Number is administered by the ASN owner
# Format is ip:route_target  (4 bytes:2 bytes)
# 0x02 and 0x03 .. read the RFC :)




# encoding: utf-8
"""
fragments.py

Created by Thomas Mangin on 2010-02-04.
Copyright (c) 2010-2012 Exa Networks. All rights reserved.
"""

class Fragments (int):
#	reserved  = 0xF0
	DONT      = 0x08
	IS        = 0x40
	FIRST     = 0x20
	LAST      = 0x10

	def __str__ (self):
		if self == 0x00:       return 'not-a-fragment'
		if self == self.DONT:  return 'dont-fragment'
		if self == self.IS:    return 'is-fragment'
		if self == self.FIRST: return 'first-fragment'
		if self == self.LAST:  return 'last-fragment'
		return 'unknown fragment value %d' % int.__str__(self)

	def __repr__ (self):
		return str(self)

def NamedFragments (name):
	fragment = name.lower()
	if fragment == 'not-a-fragment': return Fragments(0x00)
	if fragment == 'dont-fragment':  return Fragments(Fragments.DONT)
	if fragment == 'is-fragment':    return Fragments(Fragments.IS)
	if fragment == 'first-fragment': return Fragments(Fragments.FIRST)
	if fragment == 'last-fragment':  return Fragments(Fragments.LAST)
	raise ValueError('invalid fragment name %s' % fragment)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import traceback

import select

import socket

from time import sleep

from bagpipe.bgp.engine.bgp_peer_worker import BGPPeerWorker, \
    KeepAliveReceived, SendKeepAlive, FSM, InitiateConnectionException, \
    OpenWaitTimeout, StoppedException
from bagpipe.bgp.engine import RouteEvent

from bagpipe.bgp.common.looking_glass import LookingGlass

from bagpipe.exabgp.network.connection import Connection
from bagpipe.exabgp.network.protocol import Protocol, Failure
from bagpipe.exabgp.structure.neighbor import Neighbor

from bagpipe.exabgp.structure.address import AFI, SAFI

from bagpipe.exabgp.message.update.attribute.communities import RouteTarget
from bagpipe.exabgp.message.nop import NOP
from bagpipe.exabgp.message.open import Open, RouterID, Capabilities
from bagpipe.exabgp.message.update import Update
from bagpipe.exabgp.message.keepalive import KeepAlive
from bagpipe.exabgp.message.notification import Notification
from bagpipe.exabgp.message.update.route import Route
from bagpipe.exabgp.message.update.attribute.id import AttributeID


class FakePeer(object):

    '''Dummy class to be able to to plug into exabgp code'''

    def __init__(self, neighbor):
        self.afi = 1  # only IPv4 for now
        self.ip = neighbor.peer_address
        self.neighbor = neighbor


class FakeLocal(object):

    '''Dummy class to be able to to plug into exabgp code'''

    def __init__(self, ip):
        self.afi = 1  # only IPv4 for now
        self.ip = ip


class MyBGPProtocol(Protocol):

    '''Extends exabgp's Protocol class, but changes the new_open method'''

    def new_open(self, restarted, asn4, config, enabledFamilies=[]):
        '''Same as exabgp.Protocol.new_open except that we advertise support
        for MPLS VPN and RTC'''

        asn = self.neighbor.local_as
        # (we don't support ASN4)

        o = Open(4, asn, self.neighbor.router_id.ip, Capabilities()
                 .default(self.neighbor, restarted), self.neighbor.hold_time)

        o.capabilities[Capabilities.MULTIPROTOCOL_EXTENSIONS].remove(
            (AFI(AFI.ipv4), SAFI(SAFI.flow_ipv4)))
        o.capabilities[Capabilities.MULTIPROTOCOL_EXTENSIONS].remove(
            (AFI(AFI.ipv4), SAFI(SAFI.unicast)))
        o.capabilities[Capabilities.MULTIPROTOCOL_EXTENSIONS].remove(
            (AFI(AFI.ipv6), SAFI(SAFI.unicast)))
        for afi_safi in enabledFamilies:
            o.capabilities[
                Capabilities.MULTIPROTOCOL_EXTENSIONS].append(afi_safi)

        if config['enable_rtc']:
            o.capabilities[Capabilities.MULTIPROTOCOL_EXTENSIONS].append(
                (AFI(AFI.ipv4), SAFI(SAFI.rtc)))

        if not self.connection.write(o.message()):
            raise Exception("Error while sending open")

        return o


class ExaBGPPeerWorker(BGPPeerWorker, LookingGlass):

    enabledFamilies = [(AFI(AFI.ipv4), SAFI(SAFI.mpls_vpn)),
                       # (AFI(AFI.ipv6), SAFI(SAFI.mpls_vpn)),
                       (AFI(AFI.l2vpn), SAFI(SAFI.evpn))]

    def __init__(self, bgpManager, name, peerAddress, config):
        BGPPeerWorker.__init__(self, bgpManager, name, peerAddress)
        self.config = config
        self.localAddress = self.config['local_address']
        self.peerAddress = peerAddress

        self.connection = None

        self.rtc_active = False
        self._activeFamilies = []

    def _toIdle(self):
        self._activeFamilies = []

    def _initiateConnection(self):
        self.log.debug("Initiate ExaBGP connection to %s from %s",
                       self.peerAddress, self.localAddress)

        self.rtc_active = False

        neighbor = Neighbor()
        neighbor.router_id = RouterID(self.config['local_address'])
        neighbor.local_as = self.config['my_as']
        neighbor.peer_as = self.config['peer_as']
        neighbor.local_address = self.config['local_address']
        neighbor.peer_address = self.peerAddress
        neighbor.parse_routes = True

        # create dummy objects to fake exabgp into talking with us
        peer = FakePeer(neighbor)
        local = FakeLocal(self.localAddress)

        try:
            self.connection = Connection(peer, local, None, None)
        except Failure as e:
            raise InitiateConnectionException(repr(e))

        self.log.debug("Instantiate ExaBGP Protocol")
        self.protocol = MyBGPProtocol(peer, self.connection)
        self.protocol.connect()

        # this is highly similar to exabgp.network.peer._run

        o = self.protocol.new_open(
            False, False, self.config, ExaBGPPeerWorker.enabledFamilies)

        self.log.debug("Send open: [%s]", o)
        self.fsm.state = FSM.OpenSent

        count = 0
        self.log.debug("Wait for open...")
        while not self.shouldStop:
            # FIXME: we should time-out here, at some point
            message = self.protocol.read_open(o, None)

            count += 1
            if isinstance(message, NOP):
                # TODO(tmmorin): check compliance with BGP specs...
                if count > 20:
                    self.connection.close()
                    # FIXME: this should be moved to
                    # BGPPeerWorker in a more generic way
                    # (+ send Notify when needed)
                    raise OpenWaitTimeout("%ds" % int(20 * 0.5))
                sleep(0.5)
                continue

            self.log.debug("Read message: %s", message)

            if isinstance(message, Open):
                break
            else:
                self.log.error("Received unexpected message: %s", message)
                # FIXME

        if self.shouldStop:
            raise StoppedException()

        # An Open was received
        received_open = message

        self._setHoldTime(received_open.hold_time)

        # Hack to ease troubleshooting, have the real peer address appear in
        # the logs when fakerr is used
        if received_open.router_id.ip != self.peerAddress:
            self.log.info("changing thread name from %s to BGP-x%s, based on"
                          " the router-id advertized in Open (different from"
                          " peerAddress == %s)", self.name,
                          received_open.router_id.ip, self.peerAddress)
            self.name = "BGP-%s:%s" % (self.peerAddress,
                                       received_open.router_id.ip)

        try:
            mp_capabilities = received_open.capabilities[
                Capabilities.MULTIPROTOCOL_EXTENSIONS]
        except Exception:
            mp_capabilities = []

        # check that our peer advertized at least mpls_vpn and evpn
        # capabilities
        self._activeFamilies = []
        for (afi, safi) in (ExaBGPPeerWorker.enabledFamilies +
                            [(AFI(AFI.ipv4), SAFI(SAFI.rtc))]):
            if (afi, safi) not in mp_capabilities:
                self.log.warning(
                    "Peer does not advertise (%s,%s) capability", afi, safi)
            else:
                self.log.info(
                    "Family (%s,%s) successfully negotiated with peer %s",
                    afi, safi, self.peerAddress)
                self._activeFamilies.append((afi, safi))

        if len(self._activeFamilies) == 0:
            self.log.error("No family was negotiated for VPN routes")

        # proceed BGP session

        self.connection.io.setblocking(1)

        self.enqueue(SendKeepAlive)

        self.fsm.state = FSM.OpenConfirm

        self.rtc_active = False

        if self.config['enable_rtc']:
            if (AFI(AFI.ipv4), SAFI(SAFI.rtc)) in mp_capabilities:
                self.log.info(
                    "RTC successfully enabled with peer %s", self.peerAddress)
                self.rtc_active = True
            else:
                self.log.warning(
                    "enable_rtc True but peer not configured for RTC")

    def _toEstablished(self):
        BGPPeerWorker._toEstablished(self)

        if self.rtc_active:
            # subscribe to RTC routes, to be able to propagate them from
            # internal workers to this peer
            self._subscribe(AFI(AFI.ipv4), SAFI(SAFI.rtc))
        else:
            # if we don't use RTC with our peer, then we need to see events for
            # all routes of all active families, to be able to send them to him
            for (afi, safi) in self._activeFamilies:
                self._subscribe(afi, safi)

    def _receiveLoopFun(self):

        select.select([self.connection.io], [], [], 5)

        if not self._queue.empty():
            if self._stopLoops.isSet():
                self.log.info("stopLoops is set -> Close connection and"
                              " Finish receive Loop")
                self.connection.close()
                return 0

        try:
            message = self.protocol.read_message()
        except Notification as e:
            self.log.error("Peer notified us about an error: %s", e)
            return 2
        except Failure as e:
            self.log.warning("Protocol failure: %s", e)
            return 2
        except socket.error as e:
            self.log.warning("Socket error: %s", e)
            return 2
        except Exception as e:
            self.log.error("Error while reading BGP message: %s", e)
            raise

        if message.TYPE in (NOP.TYPE):
            # we arrived here because select call timed-out
            return 1
        elif message.TYPE == Update.TYPE:
            if (self.fsm.state != FSM.Established):
                raise Exception("Update received but not in Established state")
            pass  # see below
        elif message.TYPE == KeepAlive.TYPE:
            if (self.fsm.state == FSM.OpenConfirm):
                self._toEstablished()
            self.enqueue(KeepAliveReceived)
            self.log.debug("Received message: %s", message)
        else:
            self.log.warning("Received unexpected message: %s", message)

        if isinstance(message, Update):
            self.log.info("Received message: UPDATE...")
            if message.routes:
                for route in message.routes:
                    self._processReceivedRoute(route)

        return 1

    def _processReceivedRoute(self, route):
        self.log.info("Received route: %s", route)

        rts = []
        if AttributeID.EXTENDED_COMMUNITY in route.attributes:
            rts = [ecom for ecom in route.attributes[
                   AttributeID.EXTENDED_COMMUNITY].communities
                   if isinstance(ecom, RouteTarget)]

            if not rts:
                raise Exception("Unable to find any Route Targets"
                                "in the received route")

        routeEntry = self._newRouteEntry(route.nlri.afi, route.nlri.safi, rts,
                                         route.nlri, route.attributes)

        if route.action == "announce":
            self._pushEvent(RouteEvent(RouteEvent.ADVERTISE, routeEntry))
        else:
            self._pushEvent(RouteEvent(RouteEvent.WITHDRAW, routeEntry))

        # TODO(tmmorin): move RTC code out-of the peer-specific code
        if (route.nlri.afi, route.nlri.safi) == (AFI(AFI.ipv4),
                                                 SAFI(SAFI.rtc)):
            self.log.info("Received an RTC route")

            if route.nlri.route_target is None:
                self.log.info("Received RTC is a wildcard")

            # the semantic of RTC routes does not distinguish between AFI/SAFIs
            # if our peer subscribed to a Route Target, it means that we needs
            # to send him all routes of any AFI/SAFI carrying this RouteTarget.
            for (afi, safi) in self._activeFamilies:
                if (afi, safi) != (AFI(AFI.ipv4), SAFI(SAFI.rtc)):
                    if route.action == "announce":
                        self._subscribe(afi, safi, route.nlri.route_target)
                    else:  # withdraw
                        self._unsubscribe(afi, safi, route.nlri.route_target)

    def _send(self, data):
        # (error if state not the right one for sending updates)
        self.log.debug("Sending %d bytes on socket to peer %s",
                       len(data), self.peerAddress)
        try:
            self.connection.write(data)
        except Exception as e:
            self.log.error("Was not able to send data: %s", e)

    def _keepAliveMessageData(self):
        return KeepAlive().message()

    def _updateForRouteEvent(self, event):
        r = Route(event.routeEntry.nlri)
        if event.type == event.ADVERTISE:
            self.log.info("Generate UPDATE message: %s", r)
            r.attributes = event.routeEntry.attributes
            try:
                return Update([r]).update(False, self.config['my_as'],
                                          self.config['my_as'])
            except Exception as e:
                self.log.error("Exception while generating message for "
                               "route %s: %s", r, e)
                self.log.warning("%s", traceback.format_exc())
                return ''

        elif event.type == event.WITHDRAW:
            self.log.info("Generate WITHDRAW message: %s", r)
            return Update([r]).withdraw(False, self.config['my_as'],
                                        self.config['my_as'])

    def stop(self):
        if self.connection is not None:
            self.connection.close()
        BGPPeerWorker.stop(self)

    # Looking Glass ###############

    def getLookingGlassLocalInfo(self, pathPrefix):
        return {
            "peeringAddresses": {"peerAddress":  self.peerAddress,
                                 "localAddress": self.localAddress},
            "as_info": {"local": self.config['my_as'],
                        "peer":  self.config['peer_as']},
            "rtc": {"active": self.rtc_active,
                    "enabled": self.config['enable_rtc']},
            "active_families": [repr(f) for f in self._activeFamilies],
        }



# encoding: utf-8
"""
nop.py

Created by Thomas Mangin on 2009-11-06.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
"""

from bagpipe.exabgp.message import Message

class NOP (Message):
	TYPE = chr(0x00)

	def __init__ (self,data):
		self.data = data

	def message (self):
		return self._message(self.data)




# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module classes related to producing and consuming events related to
BGP routes.

routes: RouteEntry

events: RouteEvent
    an announcement or a withdrawal of a BGP route

workers: Worker
    * produce events
    * subscribe to the route table manager to consume events related to
      certain BGP routes

route table manager (singleton)
    * tracks subscriptions of workers
    * dispatches events based on subscriptions

"""

import logging

from bagpipe.exabgp.structure.address import AFI, SAFI
from bagpipe.exabgp.message.update.attribute.communities import RouteTarget
from bagpipe.exabgp.message.update.attribute import AttributeID
from bagpipe.exabgp.message.update.attributes import Attributes

from bagpipe.bgp.common.looking_glass import LookingGlass, \
    LookingGlassReferences

log = logging.getLogger(__name__)


class RouteEntry(LookingGlass):

    """A route entry describes a BGP route, i.e. the association of:

* a BGP NLRI of a specific type (e.g. a VPNv4 route
  like "1.2.3.4:5:192.168.0.5/32")
* BGP attributes
* the source of the BGP route (e.g. the BGP peer, or the local VPN instance,
  that advertizes the route)

"""

    def __init__(self, afi, safi, routeTargets, nlri, attributes, source):
        assert(isinstance(afi, AFI))
        assert(isinstance(safi, SAFI))
        assert(isinstance(attributes, Attributes))
        self.source = source
        self.afi = afi
        self.safi = safi
        self.nlri = nlri
        self.attributes = attributes
        # a list of exabgp.message.update.attribute.communities.RouteTarget:
        self.routeTargets = routeTargets

    def __cmp__(self, other):
        if (isinstance(other, RouteEntry) and
                self.afi == other.afi and
                self.safi == other.safi and
                self.source == other.source and
                self.nlri == other.nlri and
                self.attributes.sameValuesAs(other.attributes)):
            return 0
        else:
            return -1

    def __hash__(self):  # FIXME: improve for better performance ?
        return hash("%d/%d %s %d %s" % (self.afi, self.safi, self.source,
                                        hash(self.nlri), hash(self.attributes)
                                        ))

    def __repr__(self):
        fromString = " from:%s" % self.source if self.source else ""
        return "[RouteEntry: %s %s %s %s RT:%s%s]" % (self.afi, self.safi,
                                                      self.nlri,
                                                      self.attributes,
                                                      self.routeTargets,
                                                      fromString)

    def getLookingGlassLocalInfo(self, pathPrefix):

        attributesDict = {}

        for (attributeId, value) in self.attributes.iteritems():

            # skip some attributes that we care less about
            if (attributeId == AttributeID.AS_PATH or
               attributeId == AttributeID.ORIGIN or
               attributeId == AttributeID.LOCAL_PREF):
                continue

            attributesDict[
                str(AttributeID(attributeId)).lower()] = repr(value)

        res = {"afi-safi": "%s/%s" % (self.afi, self.safi),
               "attributes": attributesDict
               }

        if self.source:
            res["source"] = {"id": self.source.name,
                             "href": LookingGlassReferences.getAbsolutePath(
                                 "BGP_WORKERS", pathPrefix, [self.source.name])
                             }

        if (self.safi) in [SAFI.mpls_vpn, SAFI.evpn]:
            res["route_targets"] = [repr(rt) for rt in self.routeTargets]

        return {
            repr(self.nlri): res
        }


class RouteEvent(object):

    """A RouteEvent represents an advertisement or a withdrawal of a RouteEntry
    """

    # event Types
    ADVERTISE = 1
    WITHDRAW = 2

    type2name = {ADVERTISE: "Advertise",
                 WITHDRAW: "Withdraw"}

    def __init__(self, eventType, routeEntry, source=None):
        assert(eventType in RouteEvent.type2name.keys())
        assert(isinstance(routeEntry, RouteEntry))
        self.type = eventType
        self.routeEntry = routeEntry
        if source is not None:
            self.source = source
        else:
            self.source = routeEntry.source
        self.replacedRoute = None

    def setReplacedRoute(self, replacedRoute):
        ''' Called only by RouteTableManager, replacedRoute should be a
        RouteEntry '''
        assert(isinstance(replacedRoute, RouteEntry)
               or (replacedRoute is None))
        assert(replacedRoute != self.routeEntry)
        self.replacedRoute = replacedRoute

    def __repr__(self):
        if self.replacedRoute:
            replacesStr = "replaces one route"
        else:
            replacesStr = "replaces no route"
        return "[RouteEvent(%s): %s %s %s]" % (replacesStr,
                                               RouteEvent.type2name[self.type],
                                               self.routeEntry,
                                               self.source)


class _SubUnsubCommon(object):

    def __init__(self, afi, safi, routeTarget, worker=None):
        assert(isinstance(afi, AFI))
        assert(isinstance(safi, SAFI))
        assert(routeTarget is None or isinstance(routeTarget, RouteTarget))
        self.afi = afi
        self.safi = safi
        self.routeTarget = routeTarget
        self.worker = worker

    def __repr__(self):
        byWorker = " by %s" % self.worker.name if self.worker else ""
        return "%s [%s/%s,%s]%s" % (self.__class__.__name__,
                                    self.afi or "*", self.safi or "*",
                                    self.routeTarget or "*", byWorker)


class Subscription(_SubUnsubCommon):

    """Represents a Subscription to RouteEvents

A subscription specifies the AFI, the SAFI, and the Route Target of the
RouteEntry for which the subscriber wants to receive events.

Any of these (afi, safi or route target) can be replaced by a wildcard:

* Subscription.ANY_AFI
* Subscription.ANY_SAFI
* Subscription.ANY_RT
    """

    ANY_AFI = AFI(0)
    ANY_SAFI = SAFI(0)
    ANY_RT = None

    def __init__(self, afi, safi, routeTarget=None, worker=None):
        _SubUnsubCommon.__init__(self, afi, safi, routeTarget, worker)


class Unsubscription(_SubUnsubCommon):

    def __init__(self, afi, safi, routeTarget=None, worker=None):
        _SubUnsubCommon.__init__(self, afi, safi, routeTarget, worker)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import re

import urllib

from bagpipe.bgp.common.utils import enum

log = logging.getLogger(__name__)

LGMap = enum('VALUE', 'SUBITEM', 'SUBTREE',
             'FORWARD', 'COLLECTION', 'DELEGATE')


def _splitLGPath(pathPrefix, path):
    if len(path) == 0:
        return (None, None, pathPrefix)
    else:
        return (path[0], path[1:],
                "%s/%s" % (pathPrefix, urllib.quote(path[0])))


def _getLGLocalInfoRecurse(obj, cls, pathPrefix):

    if cls == LookingGlass:
        return {}

    log.debug("_getLGLocalInfoRecurse: %s", cls)

    result = cls.getLookingGlassLocalInfo(obj, pathPrefix)

    assert(isinstance(result, dict))

    for base in cls.__bases__:
        if issubclass(base, LookingGlass):
            result.update(
                _getLGLocalInfoRecurse(obj, base, pathPrefix))

    return result


def _getLGMapRecurse(obj, cls):

    if cls == LookingGlass:
        return {}

    log.debug("_getLGMapRecurse: %s", cls)

    result = cls.getLGMap(obj)

    for base in cls.__bases__:
        if issubclass(base, LookingGlass):
            result.update(_getLGMapRecurse(obj, base))
        else:
            log.debug("not recursing into %s", base)

    return result


def _lookupPathInDict(myDict, path):

    log.debug("_lookupPathInDict: %s vs. %s", myDict, path)

    assert(isinstance(path, list))

    if len(path) == 0:
        log.debug("path len is zero, returning myDict %s", myDict)
        return myDict

    # len(path)>0
    if not (isinstance(myDict, dict)):
        raise KeyError(path[0])
    else:
        return _lookupPathInDict(myDict[path[0]], path[1:])


class LookingGlass(object):

    def _getLGMap(self):
        """not to be overridden: calls getLGMap, on each of the super classes
           and merge the result in a dict
        """
        return _getLGMapRecurse(self, self.__class__)

    def getLGMap(self):
        """
        Can be overriden by looking glass objects to indicate looking
        glass information items for this objects.

        :returns: a dict mapping a <subpath> to a (<lg_map_type>,<hook>) tuple

        if lg_map_type is VALUE, then the looking glass information for
            <subpath> will be <hook>
        if lg_map_type is SUBITEM, then <hook> is supposed to be a function
            and the looking glass information for <subpath> will be the result
            of calling hook()
        if lg_map_type is SUBTREE, then <hook> is supposed to be a function and
            the looking glass information for <subpath> will be the result of
            calling hook(pathPrefix), but this information will only be
            produced if the <subpath> is queried (not produced if the full
             object is queried)
        if lg_map_type is FORWARD, then <hook> is supposed to be a looking
            glass object and the looking glass information for <subpath>
            will be the looking glass information for <subpath> of object
            <hook>
        if lg_map_type is DELEGATE, then <hook> is supposed to be a looking
            glass object and the looking glass information for <subpath> will
            be the full looking glass information for object <hook>
        if lg_map_type is COLLECTION, then <hook> is supposed to be a tuple of
            functions (listCallback,targetCallback). listCallback() is expected
            to return a list of string, each string identifying a looking
            glass object targetCallback(string) is expected to return the
            looking glass object corresponding to <string>

            if *self* is directly queried, the information returned is just a
            list of dict containing "href" values pointing to each object in
            the collection
            if a <subpath> is queried, the information returned is the
            looking glass information for the object corresponding to
            <subpath>
        """
        return {}

    def _getLookingGlassLocalInfo(self, pathPrefix):
        """
        not to be overridden: calls getLookingGlassLocalInfo, on each of
          the super classes and merge the result in a dict
        """
        return _getLGLocalInfoRecurse(self, self.__class__, pathPrefix)

    def getLookingGlassLocalInfo(self, pathPrefix):
        """
        Can be overriden by looking glass objects.

        :param pathPrefix: the URL prefix that was used to reach *self*
            through the looking glass
        :returns: a dict that will be serialized as JSON and passed to the
            looking glass client, either as is, or if a sub path was queried,
            the dict value corresponding to the first item of the path
        """
        return {}

    def getLookingGlassInfo(self, pathPrefix="", path=None):
        """
        This method builds the looking glass information for *self* based on
        the looking glass map

        *not* to be overridden by looking glass objects
        """
        if path is None:
            path = []

        (firstSegment, restOfPath,
         newPathPrefix) = _splitLGPath(pathPrefix, path)

        lgMap = self._getLGMap()

        if (firstSegment in lgMap):
            (mappingType, mappingTarget) = lgMap[firstSegment]
            log.debug("Delegation for pathItem '%s': %s:%s ",
                      firstSegment, LGMap.reverse[mappingType], mappingTarget)

            if mappingType == LGMap.VALUE:
                return mappingTarget

            if mappingType == LGMap.FORWARD:
                log.debug(
                    "   Forwarded '%s' to target %s...", path, mappingTarget)
                if not isinstance(mappingTarget, LookingGlass):
                    log.error(
                        "Delegation target for '%s' at '%s' does not "
                        "implement LookingGlass!", firstSegment, newPathPrefix)
                    raise NoSuchLookingGlassObject(
                        newPathPrefix, firstSegment)
                return mappingTarget.getLookingGlassInfo(pathPrefix, path)

            if mappingType == LGMap.FORWARD:
                log.debug(
                    "   Forwarded '%s' to target %s...", path, mappingTarget)
                if not isinstance(mappingTarget, LookingGlass):
                    log.error(
                        "Delegation target for '%s' at '%s' does not "
                        "implement LookingGlass!", firstSegment, newPathPrefix)
                    raise NoSuchLookingGlassObject(newPathPrefix, firstSegment)
                return mappingTarget.getLookingGlassInfo(pathPrefix, path)

            elif mappingType == LGMap.DELEGATE:
                log.debug(
                    "   Delegated '%s' to delegation target %s ...",
                    path, mappingTarget)
                if not isinstance(mappingTarget, LookingGlass):
                    log.error(
                        "Delegation target for '%s' at '%s' does not "
                        "implement LookingGlass!", firstSegment, newPathPrefix)
                    raise NoSuchLookingGlassObject(newPathPrefix, firstSegment)
                return mappingTarget.getLookingGlassInfo(newPathPrefix,
                                                         restOfPath)

            elif mappingType == LGMap.SUBITEM:
                log.debug("   Sub-item callback: %s", firstSegment)
                try:
                    return _lookupPathInDict(mappingTarget(), restOfPath)
                except KeyError as e:
                    raise NoSuchLookingGlassObject(newPathPrefix, str(e))

            elif mappingType == LGMap.SUBTREE:
                log.debug("   Subtree callback: %s(...)", firstSegment)
                try:
                    return _lookupPathInDict(mappingTarget(newPathPrefix),
                                             restOfPath)
                except KeyError as e:
                    raise NoSuchLookingGlassObject(newPathPrefix, str(e))

            elif mappingType == LGMap.COLLECTION:
                log.debug("   Collection callback...")

                (listCallback, targetCallback) = mappingTarget

                (secondSegment, restOfPath, newerPathPrefix) = \
                    _splitLGPath(newPathPrefix, restOfPath)
                if secondSegment is None:
                    log.debug("   Getting list elements: %s", listCallback)
                    result = []
                    for x in listCallback():
                        x["href"] = LookingGlass.getLGPrefixedPath(
                            pathPrefix, [firstSegment, x["id"]])
                        result.append(x)
                    return result
                else:
                    log.debug("   Callback -> resolve subItem '%s' with %s "
                              "and follow up getLookingGlassInfo(...'%s')",
                              secondSegment, targetCallback, restOfPath)
                    try:
                        target = targetCallback(secondSegment)
                        if target is None:
                            log.error("No delegation target for '%s' at '%s' ",
                                      secondSegment, newPathPrefix)
                            raise NoSuchLookingGlassObject(
                                newPathPrefix, secondSegment)
                        if not isinstance(target, LookingGlass):
                            log.error("Delegation target for '%s' at '%s' "
                                      "does not implement LookingGlass (%s)!",
                                      secondSegment, newPathPrefix,
                                      type(target))
                            raise NoSuchLookingGlassObject(
                                newPathPrefix, secondSegment)
                        return target.getLookingGlassInfo(newerPathPrefix,
                                                          restOfPath)
                    except KeyError:
                        raise NoSuchLookingGlassObject(
                            newPathPrefix, secondSegment)

        # firtSegment is None or is not in our map
        # let's build LookingGlassLocalInfo
        info = self._getLookingGlassLocalInfo(pathPrefix)
        for (pathItem, (mappingType, mappingTarget)) in lgMap.iteritems():
            if pathItem in info:
                log.warning("overriding '%s', present both in "
                            "LookingGlassLocalInfo and LookingGlass map",
                            pathItem)
            if mappingType in (LGMap.FORWARD, LGMap.DELEGATE, LGMap.SUBTREE,
                               LGMap.COLLECTION):
                info[pathItem] = {"href":
                                  LookingGlass.getLGPrefixedPath(pathPrefix,
                                                                 [pathItem])}
            elif mappingType == LGMap.SUBITEM:
                log.debug("   Subitem => callback %s(...)", mappingTarget)
                info[pathItem] = mappingTarget()
            elif mappingType == LGMap.VALUE:
                info[pathItem] = mappingTarget
            else:
                log.warning("LGMap not processed for %s", pathItem)

        if firstSegment is None:
            return info
        else:
            try:
                return _lookupPathInDict(info, path)
            except KeyError as e:
                raise NoSuchLookingGlassObject(newPathPrefix, str(e))

        log.warning("Looking glass did not found a looking-glass object for"
                    " this path...")

        return None

    @staticmethod
    def getLGPrefixedPath(pathPrefix, pathItems):
        fmt = "%s" + ('/%s' * len(pathItems))
        quotedPathItems = [urllib.quote(pathItem) for pathItem in pathItems]
        quotedPathItems.insert(0, pathPrefix)
        return fmt % tuple(quotedPathItems)


class NoSuchLookingGlassObject(Exception):

    def __init__(self, pathPrefix, path):
        Exception.__init__(self)
        assert(isinstance(pathPrefix, str))
        self.pathPrefix = pathPrefix

        assert(isinstance(path, str))
        self.path = path

    def __repr__(self):
        return "No such looking glass object: %s at %s" % (self.path,
                                                           self.pathPrefix)


class LookingGlassReferences(object):

    root = ""
    references = {}

    @staticmethod
    def setRoot(urlPrefix):
        LookingGlassReferences.root = urlPrefix

    @staticmethod
    def setReferencePath(reference, path):
        LookingGlassReferences.references[reference] = path

    @staticmethod
    def getAbsolutePath(reference, pathPrefix, path=None):
        if path is None:
            path = []
        index = pathPrefix.find(LookingGlassReferences.root)
        absoluteBaseURL = pathPrefix[:index + len(LookingGlassReferences.root)]
        return LookingGlass.getLGPrefixedPath(
            absoluteBaseURL,
            LookingGlassReferences.references[reference] + path)


class LookingGlassLogHandler(logging.Handler):

    """
    This log handler simply stores the last <maxSize> messages of importance
    above <level>. These messages can be retrieved with .getRecords().
    """

    def __init__(self, level=logging.WARNING, maxSize=100):
        logging.Handler.__init__(self, level)
        self.records = []
        self.maxSize = maxSize
        self.setFormatter(
            logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))

    def emit(self, record):
        # expand the log message now and free references to the arguments
        record.msg = record.getMessage().replace('"', "'")
        record.args = None
        self.records.insert(0, record)
        del self.records[self.maxSize:]

    def __len__(self):
        return len(self.records)

    def getRecords(self):
        return self.records

    def resetLocalLGLogs(self):
        del self.records[:]


class LookingGlassLocalLogger(LookingGlass):

    """
    For objects subclassing this class, self.log will be a logger derived from
    <name> based on the existing logging configuration, but with an additional
    logger using LookingGlassLogHandler.

    This additional logger is used to make the last <n> records (above WARNING)
    available through the looking glass
    """

    def __init__(self, appendToName=""):
        try:
            self.lgLogHandler
        except AttributeError:
            self.lgLogHandler = LookingGlassLogHandler()
            name = self.__module__  # + "." + self.__class__.__name__
            if appendToName:
                name += "." + appendToName
            elif (hasattr(self, 'instanceId')):
                name += ".%d" % self.instanceId
            elif (hasattr(self, 'name')):
                name += ".%s" % re.sub("[. ]", "-", self.name).lower()
            self.log = logging.getLogger(name)
            self.log.addHandler(self.lgLogHandler)

    def getLGMap(self):
        return {"logs": (LGMap.SUBTREE, self.getLogs)}

    def getLogs(self, pathPrefix):
        return [{'level': record.levelname,
                 'time': self.lgLogHandler.formatter.formatTime(record),
                 'message': record.msg}
                for record in self.lgLogHandler.getRecords()]

    def _resetLocalLGLogs(self):
        self.lgLogHandler.resetLocalLGLogs()



# encoding: utf-8
"""
route.py

Created by Thomas Mangin on 2010-01-16.
Copyright (c) 2010-2012 Exa Networks. All rights reserved.
"""

from bagpipe.exabgp.structure.address import Address
from bagpipe.exabgp.message.update.attributes import Attributes

# This class must be separated from the wire representation of a Route
# =================================================================== Route

class Route (object):
	def __init__ (self,nlri):
		self.nlri = nlri
		self.__address = Address(nlri.afi,nlri.safi)
		self.attributes = Attributes()

	def __str__ (self):
		return "%s %s%s" % (str(self.__address),str(self.nlri),str(self.attributes))

	def __repr__ (self):
		return str(self)

	def __eq__ (self,other):
		return str(self) == str(other)

class ReceivedRoute (Route):
	def __init__ (self,nlri,action):
		self.action = action	# announce or withdraw
		Route.__init__(self,nlri)

	def __str__ (self):
		return "%s %s" % (self.action,Route.__str__(self))



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""

.. module:: test_tracker_worker
   :synopsis: a module that defines several test cases for the tracker_worker
              module.
   In particular, unit tests for TrackerWorker class.
   Setup: Run TrackerWorker instance.
   TearDown: Stop TrackerWorker instance.
   TrackerWorker is in charge to receive RouteEvent from RouteTableManager.
   A RouteEvent contains an event type ADVERTIZE or WITHDRAW, and a RouteEntry.
   TrackerWorker should call _newBestRoute and/or _bestRouteRemoved if the new
   RouteEntry changes the current list of the known best routes. The current
   list of the known best routes, which can be modified by the new RouteEntry,
   is selected thanks to the trackedEntry associated to the new RouteEntry.
   The trackedEntry is obtained thanks to _route2TrackedEntry.
   _compareRoutes is used to compare 2 RouteEntry.
   Unit tests are organized as follow:
   TestA: basic tests, advertise several routes with different NLRI and same or
          different sources
   TestB: same routes (with _compareRoutes) announced by different sources
   TestC: different routes (with _compareRoutes) announced by different
          sources, TrackerWorker selects the best route.
   TestD: ECMP routes or same routes (with _compareRoutes), same source, same
          attributes except NextHop
   TestE: different routes (with compareRoutes announced by the same source
          with replacedRoute not none
"""
import mock

from copy import copy

from testtools import TestCase
from threading import Thread

from bagpipe.bgp.tests import BaseTestBagPipeBGP, RT1, RT2, NLRI1, NLRI2, \
    NH1, NH2, NH3, NBR, BRR
from bagpipe.bgp.engine import RouteEvent
from bagpipe.bgp.engine.worker import Worker
from bagpipe.bgp.engine.tracker_worker import TrackerWorker
from bagpipe.exabgp.message.update.attribute import AttributeID

import logging

log = logging.getLogger()


def _test_compareRoutes(self, routeA, routeB):
    if (routeA.nlri != routeB.nlri or
            routeA.afi != routeB.afi or
            routeA.safi != routeB.safi):
        raise Exception('Bug: compareRoutes called with routes having '
                        'different nlri/afi/safi')
    else:
        if (routeA.attributes.sameValuesAs(routeB.attributes)):
            return 0
        else:
            lpA = routeA.attributes[AttributeID.LOCAL_PREF].localpref
            nhA = routeA.attributes[AttributeID.NEXT_HOP].next_hop

            lpB = routeB.attributes[AttributeID.LOCAL_PREF].localpref
            nhB = routeB.attributes[AttributeID.NEXT_HOP].next_hop

            if nhA != nhB and lpA == lpB:
                    # ECMP routes
                return 0
            else:
                return cmp(lpA, lpB)


class TrackerWorkerThread(TrackerWorker, Thread):

    def __init__(self):
        Thread.__init__(self, name='TrackerWorkerThread')
        self.setDaemon(True)
        TrackerWorker.__init__(
            self, 'BGPManager', 'TrackerWorker', _test_compareRoutes)

    def stop(self):
        self._pleaseStop.set()
        self._queue.put(self.stopEvent)
        self._stopped()

    def _route2trackedEntry(self, route):
        return route.nlri

    # the definitions below are needed because TrackerWorker is an abstract
    # class
    def _newBestRoute(self, entry, route):
        pass

    def _bestRouteRemoved(self, entry, route):
        pass


class TestTrackerWorker(TestCase, BaseTestBagPipeBGP):

    def setUp(self):
        super(TestTrackerWorker, self).setUp()
        self.trackerWorker = TrackerWorkerThread()
        self.trackerWorker.start()
        self.setEventTargetWorker(self.trackerWorker)
        self._calls = []

    def tearDown(self):
        super(TestTrackerWorker, self).tearDown()
        self.trackerWorker.stop()
        self.trackerWorker.join()

    def _checkCalls(self, call_args_list, expected_list):
        for ((callArgs, _), expected) in zip(call_args_list, expected_list):
            self.assertEquals(expected[0], callArgs[0], 'Bad prefix')

            observedRouteEntry = copy(callArgs[1])
            observedRouteEntry.source = None
            expectedRouteEntry = copy(expected[1])
            expectedRouteEntry.source = None
            self.assertEquals(expectedRouteEntry, observedRouteEntry,
                              "bad route Entry")

            if len(expected) >= 3:
                self.assertEquals(expected[2], callArgs[2], 'wrong last flag')

    def _callList(self, method):
        def side_effect(*args, **kwargs):
            self._append_call(method)
        return side_effect

    def testA1_differentNLRISameSource(self):
        # A source A advertises and withdraws routes for different NLRI.
        # Mock objects
        self.trackerWorker._newBestRoute = mock.Mock()
        self.trackerWorker._bestRouteRemoved = mock.Mock()

        # Only 1 source A
        workerA = Worker('BGPManager', 'Worker-A')
        # Source A advertises a route for NLRI1
        routeNlri1A = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerA, NH1, 100)
        # Source A advertises a route for NLRI2
        routeNlri2A = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI2, [RT1, RT2], workerA, NH1, 100)
        # Source A withdraws the route for NLRI1
        self._newRouteEvent(
            RouteEvent.WITHDRAW, NLRI1, [RT1, RT2], workerA, NH1, 100)
        # Source A withdraws the route for NLRI2
        self._newRouteEvent(
            RouteEvent.WITHDRAW, NLRI2, [RT1, RT2], workerA, NH1, 100)

        # Check calls and arguments list to _newBestRoute and _bestRouteRemoved
        self.assertEqual(2, self.trackerWorker._newBestRoute.call_count,
                         '2 new best routes: 1 for NLRI1 and 1 for NLRI2')
        self._checkCalls(self.trackerWorker._newBestRoute.call_args_list,
                         [(NLRI1, routeNlri1A.routeEntry),
                          (NLRI2, routeNlri2A.routeEntry)])
        self.assertEqual(2, self.trackerWorker._bestRouteRemoved.call_count,
                         '2 old routes removed: 1 for NLRI1 and 1 for NLRI2')
        self._checkCalls(
            self.trackerWorker._bestRouteRemoved.call_args_list,
            [(NLRI1, routeNlri1A.routeEntry, True),
             (NLRI2, routeNlri2A.routeEntry, True)])

    def testA2_differentNLRIDifferentSource(self):
        # 2 sources A and B advertise and withdraw routes for different NLRI.
        # Mock objects
        self.trackerWorker._newBestRoute = mock.Mock()
        self.trackerWorker._bestRouteRemoved = mock.Mock()

        # 2 sources: A and B
        workerA = Worker('BGPManager', 'Worker-A')
        workerB = Worker('BGPManager', 'Worker-B')
        # Source A advertises a route for NLRI1
        routeNlri1A = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerA, NH1, 100)
        # Source B advertises a route for NLRI2
        routeNlri2B = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI2, [RT1, RT2], workerB, NH1, 100)
        # Source A withdraws the route for NLRI1
        self._newRouteEvent(
            RouteEvent.WITHDRAW, NLRI1, [RT1, RT2], workerA, NH1, 100)
        # Source B withdraws the route for NLRI2
        self._newRouteEvent(
            RouteEvent.WITHDRAW, NLRI2, [RT1, RT2], workerB, NH1, 100)

        # Check calls and arguments list to _newBestRoute and _bestRouteRemoved
        self.assertEqual(2, self.trackerWorker._newBestRoute.call_count,
                         '2 newBestRoute calls: 1 for NLRI1 and 1 for NLRI2')
        self._checkCalls(self.trackerWorker._newBestRoute.call_args_list,
                         [(NLRI1, routeNlri1A.routeEntry),
                          (NLRI2, routeNlri2B.routeEntry)])
        self.assertEqual(2, self.trackerWorker._bestRouteRemoved.call_count,
                         '2 bestRouteRemoved calls: 1 for NLRI1 and 1 for '
                         'NLRI2')
        self._checkCalls(
            self.trackerWorker._bestRouteRemoved.call_args_list,
            [(NLRI1, routeNlri1A.routeEntry, True),
             (NLRI2, routeNlri2B.routeEntry, True)])

    def testA3_sameNLRISameSource(self):
        # A source A advertises the same route for the same NLRI
        # Mock objects
        self.trackerWorker._newBestRoute = mock.Mock()
        self.trackerWorker._bestRouteRemoved = mock.Mock()

        # 1 source: A
        workerA = Worker('BGPManager', 'Worker-A')
        # Source A advertises a route for NLRI1
        routeNlri1A = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerA, NH1, 100)
        # Source A advertises the same route for NLRI1
        self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerA, NH1, 100)

        # Check calls and arguments list to _newBestRoute and _bestRouteRemoved
        self.assertEqual(1, self.trackerWorker._newBestRoute.call_count,
                         'expected 1 newBestRoute call for NLRI1')
        self._checkCalls(self.trackerWorker._newBestRoute.call_args_list,
                         [(NLRI1, routeNlri1A.routeEntry),
                          (NLRI1, routeNlri1A.routeEntry)])

    def testA4_withdrawNLRINotKnown(self):
        # A source A withdraws a route that does not exist.
        self.trackerWorker._newBestRoute = mock.Mock()
        self.trackerWorker._bestRouteRemoved = mock.Mock()

        # 1 source: A
        workerA = Worker('BGPManager', 'Worker-A')
        # Source A withdraws a route for NLRI1 which is not known by
        # trackerWorker
        self._newRouteEvent(
            RouteEvent.WITHDRAW, NLRI1, [RT1, RT2], workerA, NH1, 100)

        # Check calls to _newBestRoute and _bestRouteRemoved
        self.assertEqual(0, self.trackerWorker._newBestRoute.call_count,
                         'newBestRoute should not have been called')
        self.assertEqual(0, self.trackerWorker._bestRouteRemoved.call_count,
                         'bestRouteRemoved should not have been called')

    def testB1_isTheCurrentBestRoute(self):
        # The route which is advertised by another source is the current best
        # route
        self.trackerWorker._newBestRoute = mock.Mock(
            side_effect=self._callList(NBR))
        self.trackerWorker._bestRouteRemoved = mock.Mock(
            side_effect=self._callList(BRR))

        # 2 sources: A and B
        workerA = Worker('BGPManager', 'Worker-A')
        workerB = Worker('BGPManager', 'Worker-B')

        # Source A advertises a route for NLRI1
        self._append_call("RE1")
        routeNlri1A = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerA, NH1, 100)
        # Source B advertises the same route for NLRI1
        self._append_call("RE2")
        routeNlri1B = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerB, NH1, 100)
        # Source A withdraws the route for NLRI1
        self._append_call("RE3")
        self._newRouteEvent(
            RouteEvent.WITHDRAW, NLRI1, [RT1, RT2], workerA, NH1, 100)
        # Source B withdraws the route for NLRI1
        self._append_call("RE4")
        self._newRouteEvent(
            RouteEvent.WITHDRAW, NLRI1, [RT1, RT2], workerB, NH1, 100)

        # Check calls and arguments list to _newBestRoute and _bestRouteRemoved
        self.assertEqual(
            1, self.trackerWorker._newBestRoute.call_count,
            '1 new best route call for NLRI1')
        self._checkCalls(
            self.trackerWorker._newBestRoute.call_args_list,
            [(NLRI1, routeNlri1A.routeEntry)])
        self.assertEqual(
            1, self.trackerWorker._bestRouteRemoved.call_count,
            '1 bestRouteRemoved call for NLRI1')
        self._checkCalls(
            self.trackerWorker._bestRouteRemoved.call_args_list,
            [(NLRI1, routeNlri1B.routeEntry, True)])

        expectedCalls = ["RE1", NBR, "RE2", "RE3", "RE4", BRR]
        self.assertEqual(expectedCalls, self._calls, 'Wrong call sequence')

    def testB2_isNotTheCurrentBestRoute(self):
        # The route which is advertised by an other source is not the current
        # best route but will become the best route
        self.trackerWorker._newBestRoute = mock.Mock(
            side_effect=self._callList(NBR))
        self.trackerWorker._bestRouteRemoved = mock.Mock(
            side_effect=self._callList(BRR))

        # 3 sources: A, B and C
        workerA = Worker('BGPManager', 'Worker-A')
        workerB = Worker('BGPManager', 'Worker-B')
        workerC = Worker('BGPManager', 'Worker-C')

        # Source A advertises route1 for NLRI1
        self._append_call("RE1")
        route1Nlri1 = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerA, NH1, 300)
        # Source B advertises route2 for NLRI1 : route1 is better than route2
        self._append_call("RE2")
        route2Nlri1 = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerB, NH1, 200)
        # Source C advertises also route2
        self._append_call("RE3")
        self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerC, NH1, 200)
        # Source A withdraws route1
        self._append_call("RE4")
        self._newRouteEvent(
            RouteEvent.WITHDRAW, NLRI1, [RT1, RT2], workerA, NH1, 300)

        # Check calls and arguments list to _newBestRoute and _bestRouteRemoved
        expectedCalls = ["RE1", NBR, "RE2", "RE3", "RE4", NBR, BRR]
        self.assertEqual(expectedCalls, self._calls, 'Wrong call sequence')

        self.assertEqual(
            2, self.trackerWorker._newBestRoute.call_count,
            '2 new best route call for NLRI1')
        self._checkCalls(self.trackerWorker._newBestRoute.call_args_list,
                         [(NLRI1, route1Nlri1.routeEntry),
                          (NLRI1, route2Nlri1.routeEntry)])
        self.assertEqual(
            1, self.trackerWorker._bestRouteRemoved.call_count,
            '1 bestRouteRemoved call for NLRI1')
        self._checkCalls(
            self.trackerWorker._bestRouteRemoved.call_args_list,
            [(NLRI1, route1Nlri1.routeEntry, False)])

    def testC1_route1BestRoute(self):
        # Route1 is the best route
        # Mock objects
        self.trackerWorker._newBestRoute = mock.Mock(
            side_effect=self._callList(NBR))
        self.trackerWorker._bestRouteRemoved = mock.Mock(
            side_effect=self._callList(BRR))

        # 2 sources : A and B
        workerA = Worker('BGPManager', 'Worker-A')
        workerB = Worker('BGPManager', 'Worker-B')

        # Source A advertises a route1 for NLRI1
        self._append_call("RE1")
        route1Nlri1A = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerA, NH1, 300)
        # Source B advertises a route2 for NLRI1 with different attributes.
        # Route1 is better than Route2
        self._append_call("RE2")
        route2Nrli1B = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerB, NH1, 200)
        # Source A withdraws route1 for NLRI1
        self._append_call("RE3")
        self._newRouteEvent(
            RouteEvent.WITHDRAW, NLRI1, [RT1, RT2], workerA, NH1, 300)
        # Source B withdraws route2 for NLRI1
        self._append_call("RE4")
        self._newRouteEvent(
            RouteEvent.WITHDRAW, NLRI1, [RT1, RT2], workerB, NH1, 200)

        # Check calls and arguments list to _newBestRoute and _bestRouteRemoved
        expectedCalls = ["RE1", NBR, "RE2", "RE3", NBR, BRR, "RE4", BRR]
        self.assertEqual(expectedCalls, self._calls, 'Wrong call sequence')

        self.assertEqual(
            2, self.trackerWorker._newBestRoute.call_count,
            '2 new newBestRoute calls for NLRI1')
        self._checkCalls(self.trackerWorker._newBestRoute.call_args_list,
                         [(NLRI1, route1Nlri1A.routeEntry),
                          (NLRI1, route2Nrli1B.routeEntry)])
        self.assertEqual(
            2, self.trackerWorker._bestRouteRemoved.call_count,
            '2 bestRouteRemoved calls for NLRI1')
        self._checkCalls(
            self.trackerWorker._bestRouteRemoved.call_args_list,
            [(NLRI1, route1Nlri1A.routeEntry, False),
             (NLRI1, route2Nrli1B.routeEntry, True)])

    def testC2_route2BestRoute(self):
        # Route2 is the best route
        # Mock objects
        self.trackerWorker._newBestRoute = mock.Mock(
            side_effect=self._callList(NBR))
        self.trackerWorker._bestRouteRemoved = mock.Mock(
            side_effect=self._callList(BRR))

        # 2 sources: A and B
        workerA = Worker('BGPManager', 'Worker-A')
        workerB = Worker('BGPManager', 'Worker-B')

        # Source A advertises a route1 for NLRI1
        self._append_call("RE1")
        route1Nlri1A = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerA, NH1, 100)
        # Source B advertises a route2 for NLRI1. Route2 is better than Route1
        self._append_call("RE2")
        route2Nrli1B = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerB, NH1, 200)
        # Source A withdraws route1 for NLRI1
        self._append_call("RE3")
        self._newRouteEvent(
            RouteEvent.WITHDRAW, NLRI1, [RT1, RT2], workerA, NH1, 100)

        # Check calls and arguments list to _newBestRoute and _bestRouteRemoved
        expectedCalls = ["RE1", NBR, "RE2", NBR, BRR, "RE3"]
        self.assertEqual(expectedCalls, self._calls, 'Wrong call sequence')

        self.assertEqual(
            2, self.trackerWorker._newBestRoute.call_count,
            '2 new newBestRoute calls for NLRI1')
        self._checkCalls(self.trackerWorker._newBestRoute.call_args_list,
                         [(NLRI1, route1Nlri1A.routeEntry),
                          (NLRI1, route2Nrli1B.routeEntry)])
        self.assertEqual(
            1, self.trackerWorker._bestRouteRemoved.call_count,
            '1 bestRouteRemoved call for NLRI1')
        self._checkCalls(
            self.trackerWorker._bestRouteRemoved.call_args_list,
            [(NLRI1, route1Nlri1A.routeEntry, False)])

    def testC3_selectNewBestRouteAmongSeveral(self):
        # When current best route is withdrawn, the new best route should be
        # selected among several routes
        self.trackerWorker._newBestRoute = mock.Mock(
            side_effect=self._callList(NBR))
        self.trackerWorker._bestRouteRemoved = mock.Mock(
            side_effect=self._callList(BRR))

        # 3 sources: A, B and C
        workerA = Worker('BGPManager', 'Worker-A')
        workerB = Worker('BGPManager', 'Worker-B')
        workerC = Worker('BGPManager', 'Worker-C')

        # Source A advertises a route1 for NLRI1
        self._append_call("RE1")
        route1Nlri1A = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerA, NH1, 300)
        # Source B advertises a route2 for NLRI1. Route1 is better than Route2
        self._append_call("RE2")
        route2Nrli1B = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerB, NH1, 200)
        # Source C advertises a route3 for NLRI1. Route2 is better than Route3
        self._append_call("RE3")
        route3Nrli1C = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerC, NH1, 100)
        # Source A withdraws route1 for NLRI1
        self._append_call("RE4")
        self._newRouteEvent(
            RouteEvent.WITHDRAW, NLRI1, [RT1, RT2], workerA, NH1, 300)
        # Source B withdraws route2 for NLRI1
        self._append_call("RE5")
        self._newRouteEvent(
            RouteEvent.WITHDRAW, NLRI1, [RT1, RT2], workerB, NH1, 200)
        # Source C withdraws route3 for NLRI1
        self._append_call("RE6")
        self._newRouteEvent(
            RouteEvent.WITHDRAW, NLRI1, [RT1, RT2], workerC, NH1, 100)

        # Check calls and arguments list to _newBestRoute and _bestRouteRemoved
        expectedCalls = ["RE1", NBR, "RE2", "RE3",
                         "RE4", NBR, BRR, "RE5", NBR, BRR, "RE6", BRR]
        self.assertEqual(expectedCalls, self._calls, 'Wrong call sequence')

        self.assertEqual(
            3, self.trackerWorker._newBestRoute.call_count,
            '3 new newBestRoute calls for NLRI1')
        self._checkCalls(self.trackerWorker._newBestRoute.call_args_list,
                         [(NLRI1, route1Nlri1A.routeEntry),
                          (NLRI1, route2Nrli1B.routeEntry),
                             (NLRI1, route3Nrli1C.routeEntry)])
        self.assertEqual(
            3, self.trackerWorker._bestRouteRemoved.call_count,
            '3 bestRouteRemoved calls for NLRI1')
        self._checkCalls(
            self.trackerWorker._bestRouteRemoved.call_args_list,
            [(NLRI1, route1Nlri1A.routeEntry, False),
             (NLRI1, route2Nrli1B.routeEntry, False),
             (NLRI1, route3Nrli1C.routeEntry, True)])

    def testD1_ECMPRoutes(self):
        # ECMP routes are routes advertised by the same worker with the same
        # LP and different NH
        self.trackerWorker._newBestRoute = mock.Mock(
            side_effect=self._callList(NBR))
        self.trackerWorker._bestRouteRemoved = mock.Mock(
            side_effect=self._callList(BRR))

        # 1 source: A
        workerA = Worker('BGPManager', 'Worker-A')

        # Source A advertises a route1 for NLRI1
        self._append_call("RE1")
        route1Nlri1A = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerA, NH1, 100)
        # Source A advertises a route2 for NLRI1. route2 is equal to route1
        # with compareRoutes, but the next_hop are different
        self._append_call("RE2")
        route2Nrli1A = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerA, NH2, 100)
        # Source A withdraws route1 for NLRI1
        self._append_call("RE3")
        self._newRouteEvent(
            RouteEvent.WITHDRAW, NLRI1, [RT1, RT2], workerA, NH1, 100)
        # Source A withdraws route2 for NLRI1
        self._append_call("RE4")
        self._newRouteEvent(
            RouteEvent.WITHDRAW, NLRI1, [RT1, RT2], workerA, NH2, 100)

        # Check calls and arguments list to _newBestRoute and _bestRouteRemoved
        expectedCalls = ["RE1", NBR, "RE2", NBR, "RE3", BRR, "RE4", BRR]
        self.assertEqual(expectedCalls, self._calls, 'Wrong call sequence')

        self.assertEqual(
            2, self.trackerWorker._newBestRoute.call_count,
            '2 new newBestRoute calls for NLRI1')
        self._checkCalls(self.trackerWorker._newBestRoute.call_args_list,
                         [(NLRI1, route1Nlri1A.routeEntry),
                          (NLRI1, route2Nrli1A.routeEntry)])
        self.assertEqual(
            2, self.trackerWorker._bestRouteRemoved.call_count,
            '2 bestRouteRemoved calls for NLRI1')
        self._checkCalls(
            self.trackerWorker._bestRouteRemoved.call_args_list,
            [(NLRI1, route1Nlri1A.routeEntry, False),
             (NLRI1, route2Nrli1A.routeEntry, True)])

    def testE1_replaceBRisNBR(self):
        # Advertise a route that replaces the best route and becomes the new
        # best route
        self.trackerWorker._newBestRoute = mock.Mock(
            side_effect=self._callList(NBR))
        self.trackerWorker._bestRouteRemoved = mock.Mock(
            side_effect=self._callList(BRR))

        # 1 source: A
        workerA = Worker('BGPManager', 'Worker-A')

        # Source A advertises a route1 for NLRI1
        self._append_call("RE1")
        route1Nlri1A = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerA, NH1, 200)
        # Source A advertises a route2 for NLRI1. Route1 is better than Route2
        # BUT Route2 replaces Route1
        self._append_call("RE2")
        route2Nrli1A = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2],
            workerA, NH1, 100, route1Nlri1A.routeEntry)

        # Check calls and arguments list to _newBestRoute and _bestRouteRemoved
        expectedCalls = ["RE1", NBR, "RE2", NBR, BRR]
        self.assertEqual(expectedCalls, self._calls, 'Wrong call sequence')

        self.assertEqual(
            2, self.trackerWorker._newBestRoute.call_count,
            '2 new newBestRoute calls for NLRI1')
        self._checkCalls(self.trackerWorker._newBestRoute.call_args_list,
                         [(NLRI1, route1Nlri1A.routeEntry),
                          (NLRI1, route2Nrli1A.routeEntry)])
        self.assertEqual(
            1, self.trackerWorker._bestRouteRemoved.call_count,
            '1 bestRouteRemoved call for NLRI1')
        self._checkCalls(
            self.trackerWorker._bestRouteRemoved.call_args_list,
            [(NLRI1, route1Nlri1A.routeEntry, False)])

    def testE2_replaceBRisNotNBR(self):
        # Advertise a route that replaces the best route but does not become
        # the new best route
        self.trackerWorker._newBestRoute = mock.Mock(
            side_effect=self._callList(NBR))
        self.trackerWorker._bestRouteRemoved = mock.Mock(
            side_effect=self._callList(BRR))

        # 2 sources : A and B
        workerA = Worker('BGPManager', 'Worker-A')
        workerB = Worker('BGPManager', 'Worker-B')

        # Source A advertises a route1 for NLRI1
        self._append_call("RE1")
        route1Nlri1A = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerA, NH1, 300)
        # Source B advertises a route2. Route1 is better than Route2
        self._append_call("RE2")
        route2Nrli1B = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerB, NH1, 200)
        # Source A advertises a route3 for NLRI1. Route3 replaces Route1.
        # Route2 is better than route3.
        self._append_call("RE3")
        route3Nrli1A = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2],
            workerA, NH1, 100, route1Nlri1A.routeEntry)
        # Source B withdraws route2 for NLRI1
        self._append_call("RE4")
        self._newRouteEvent(
            RouteEvent.WITHDRAW, NLRI1, [RT1, RT2], workerB, NH1, 200)

        # Check calls and arguments list to _newBestRoute and _bestRouteRemoved
        expectedCalls = ["RE1", NBR, "RE2", "RE3", NBR, BRR, "RE4", NBR, BRR]
        self.assertEqual(expectedCalls, self._calls, 'Wrong call sequence')

        self.assertEqual(
            3, self.trackerWorker._newBestRoute.call_count,
            '3 new newBestRoute calls for NLRI1')
        self._checkCalls(self.trackerWorker._newBestRoute.call_args_list,
                         [(NLRI1, route1Nlri1A.routeEntry),
                          (NLRI1, route2Nrli1B.routeEntry),
                             (NLRI1, route3Nrli1A.routeEntry)])
        self.assertEqual(
            2, self.trackerWorker._bestRouteRemoved.call_count,
            '2 bestRouteRemoved calls for NLRI1')
        self._checkCalls(
            self.trackerWorker._bestRouteRemoved.call_args_list,
            [(NLRI1, route1Nlri1A.routeEntry, False),
             (NLRI1, route2Nrli1B.routeEntry, False)])

    def testE3_replaceBRisNotNBR(self):
        # Advertise a route that replaces the best route but does not become
        # the new best route
        self.trackerWorker._newBestRoute = mock.Mock(
            side_effect=self._callList(NBR))
        self.trackerWorker._bestRouteRemoved = mock.Mock(
            side_effect=self._callList(BRR))

        # 3 sources: A, B and C
        workerA = Worker('BGPManager', 'Worker-A')
        workerB = Worker('BGPManager', 'Worker-B')
        workerC = Worker('BGPManager', 'Worker-C')

        # Source A advertises route1 for NLRI1
        self._append_call("RE1")
        route1Nlri1 = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerA, NH1, 300)
        # Source B advertises route2 for NLRI1 : route1 is better than route2
        self._append_call("RE2")
        route2Nlri1 = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerB, NH1, 200)
        # Source C advertises also route2
        self._append_call("RE3")
        self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerC, NH1, 200)
        # Source A advertises route3 which replaces route1
        self._append_call("RE4")
        self._newRouteEvent(RouteEvent.ADVERTISE, NLRI1, [RT1, RT2],
                            workerA, NH1, 100, route1Nlri1.routeEntry)

        # Check calls and arguments list to _newBestRoute and _bestRouteRemoved
        expectedCalls = ["RE1", NBR, "RE2", "RE3", "RE4", NBR, BRR]
        self.assertEqual(expectedCalls, self._calls, 'Wrong call sequence')

        self.assertEqual(
            2, self.trackerWorker._newBestRoute.call_count,
            '2 new best route call for NLRI1')
        self._checkCalls(self.trackerWorker._newBestRoute.call_args_list,
                         [(NLRI1, route1Nlri1.routeEntry),
                          (NLRI1, route2Nlri1.routeEntry)])
        self.assertEqual(
            1, self.trackerWorker._bestRouteRemoved.call_count,
            '1 bestRouteRemoved call for NLRI1')
        self._checkCalls(
            self.trackerWorker._bestRouteRemoved.call_args_list,
            [(NLRI1, route1Nlri1.routeEntry)])

    def testE4_notReplaceBR(self):
        # Advertise a route that does not replaces the best route and becomes
        # the new best route when the best route is withdrawn
        self.trackerWorker._newBestRoute = mock.Mock(
            side_effect=self._callList(NBR))
        self.trackerWorker._bestRouteRemoved = mock.Mock(
            side_effect=self._callList(BRR))

        # 2 sources : A and B
        workerA = Worker('BGPManager', 'Worker-A')
        workerB = Worker('BGPManager', 'Worker-B')

        # Source A advertises a route1 for NLRI1
        self._append_call("RE1")
        route1Nlri1A = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerA, NH1, 300)
        # Source B advertises a route2. Route1 is better than Route2
        self._append_call("RE2")
        route2Nrli1B = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerB, NH1, 200)
        # Source B advertises a route3 for NLRI1. Route3 replaces Route2.
        # Route1 is better than Route3
        self._append_call("RE3")
        route3Nrli1B = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2],
            workerB, NH1, 100, route2Nrli1B.routeEntry)
        # Source A withdraws route1 for NLRI1
        self._append_call("RE4")
        self._newRouteEvent(
            RouteEvent.WITHDRAW, NLRI1, [RT1, RT2], workerA, NH1, 300)

        # Check calls and arguments list to _newBestRoute and _bestRouteRemoved
        expectedCalls = ["RE1", NBR, "RE2", "RE3", "RE4", NBR, BRR]
        self.assertEqual(expectedCalls, self._calls, 'Wrong call sequence')

        self.assertEqual(
            2, self.trackerWorker._newBestRoute.call_count,
            '2 new newBestRoute calls for NLRI1')
        self._checkCalls(self.trackerWorker._newBestRoute.call_args_list,
                         [(NLRI1, route1Nlri1A.routeEntry),
                          (NLRI1, route3Nrli1B.routeEntry)])
        self.assertEqual(
            1, self.trackerWorker._bestRouteRemoved.call_count,
            '1 bestRouteRemoved call for NLRI1')
        self._checkCalls(
            self.trackerWorker._bestRouteRemoved.call_args_list,
            [(NLRI1, route1Nlri1A.routeEntry, False)])

    def testE5_replaceBRisNBREqual(self):
        # Same as E3, but the route that replaces our current best compares
        # equally to the two initially less preferred routes, and becomes best
        # route with them
        self.trackerWorker._newBestRoute = mock.Mock(
            side_effect=self._callList(NBR))
        self.trackerWorker._bestRouteRemoved = mock.Mock(
            side_effect=self._callList(BRR))

        # 3 sources: A, B and C
        workerA = Worker('BGPManager', 'Worker-A')
        workerB = Worker('BGPManager', 'Worker-B')
        workerC = Worker('BGPManager', 'Worker-C')

        # Source A advertises route1 for NLRI1
        self._append_call("RE1")
        route1 = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerA, NH1, 300)
        # Source B advertises route2 for NLRI1 : route1 is better than route2
        self._append_call("RE2")
        route2 = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerB, NH1, 200)
        # Source C advertises also route2
        self._append_call("RE3")
        route3 = self._newRouteEvent(
            RouteEvent.ADVERTISE, NLRI1, [RT1, RT2], workerC, NH2, 200)
        # Source A advertises route3 which replaces route1
        self._append_call("RE4")
        route4 = self._newRouteEvent(RouteEvent.ADVERTISE, NLRI1, [RT1, RT2],
                                     workerA, NH3, 200, route1.routeEntry)

        # Check calls and arguments list to _newBestRoute and _bestRouteRemoved
        expectedCalls = ["RE1", NBR, "RE2", "RE3", "RE4", NBR, NBR, NBR, BRR]
        self.assertEqual(expectedCalls, self._calls, 'Wrong call sequence')

        self._checkCalls(self.trackerWorker._newBestRoute.call_args_list,
                         [(NLRI1, route1.routeEntry),
                          (NLRI1, route2.routeEntry),
                             (NLRI1, route3.routeEntry),
                             (NLRI1, route4.routeEntry)])
        # FIXME: the order of route2, route3, route4 is not important in the
        # test above, we should test independently of the order

        self._checkCalls(
            self.trackerWorker._bestRouteRemoved.call_args_list,
            [(NLRI1, route1.routeEntry, False)])



# encoding: utf-8
"""
mprnlri.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
Modified by Orange - 2014
"""

from bagpipe.exabgp.structure.address import Address, SAFI #,AFI
from bagpipe.exabgp.message.update.attribute import AttributeID,Flag,Attribute

# =================================================================== MP Reacheable NLRI (15)


class MPRNLRI (Attribute):
	FLAG = Flag.OPTIONAL
	ID = AttributeID.MP_REACH_NLRI
	MULTIPLE = True

	def __init__ (self,routes):
		# all the routes must have the same next-hop
		self.routes = routes

	def pack (self):
		next_hop = ''
		# EOR do not have any next_hop
		if self.routes[0].attributes.has(AttributeID.NEXT_HOP):
			# we do not want a next_hop attribute packed (with the _attribute()) but just the next_hop itself
			next_hop = self.routes[0].attributes[AttributeID.NEXT_HOP].next_hop.pack()
			
		# FIX: for SAFI 128, the next-hop is encoded like this:
		if self.routes[0].nlri.safi == SAFI.mpls_vpn:
			next_hop = "\0"*8 + next_hop
			
		routes = ''.join([route.nlri.pack() for route in self.routes])

		return self._attribute(
			self.routes[0].nlri.afi.pack() + self.routes[0].nlri.safi.pack() +
			chr(len(next_hop)) + next_hop +
			chr(0) + routes
		)

	def __len__ (self):
		return len(self.pack())

	def __str__ (self):
		return "MP_REACH_NLRI Family %s %d NLRI(s)" % (Address.__str__(self.routes[0]),len(self.routes))

	def __repr__ (self):
		return str(self)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import subprocess


def runCommand(log, command, raiseExceptionOnError=True,
               acceptableReturnCodes=[0]):
    '''
    Executes 'command' in a subshell.
    Returns (command_output,exitCode)
        - command_output is the list of lines output on stdout by the command
    Raises an exception based on the following:
        - will only raise an Exception if raiseExceptionOnError optional
          parameter is True
        - the exit code is acceptable
        - exit code is acceptable by default if it is zero
        - exit code is acceptable if it is in the (optional)
          acceptableReturnCodes list parameter
        - putting -1 in the acceptableReturnCodes list means that *any* exit
        code is acceptable
    '''
    log.info("Running command: %s   [raiseExceptionOnError:%s]",
             command, raiseExceptionOnError)
    process = subprocess.Popen(
        command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    # Poll process for new output until finished
    output = []
    while True:
        nextline = process.stdout.readline().strip("\n")
        if nextline == '' and process.poll() is not None:
            break
        if nextline != '':
            log.debug("runCommand output: %s", nextline.rstrip())
            output.append(nextline)

    exitCode = process.returncode

    if (exitCode in acceptableReturnCodes) or (-1 in acceptableReturnCodes):
        return (output, exitCode)
    else:
        if len(output) > 0:
            message = \
                "Exit code %d when running '%s': %s" % (exitCode, command,
                                                        output[-1])
        else:
            message = \
                "Exit code %d when running '%s' (no output)" % (exitCode,
                                                                command)

        if raiseExceptionOnError:
            log.error(message)
            raise Exception(message)
        else:
            log.warning(message)
            return (output, exitCode)



# encoding: utf-8
"""
__init__.py

Created by Thomas Mangin on 2010-01-15.
Copyright (c) 2010-2012 Exa Networks. All rights reserved.
"""

from struct import pack,unpack

# We do not implement the RFC State Machine so .. we do not care :D
class State (object):
	IDLE        = 0x01
	CONNECT     = 0x02
	ACTIVE      = 0x03
	OPENSENT    = 0x04
	OPENCONFIRM = 0x05
	ESTABLISHED = 0x06


# README: the name is HORRIBLE, fix this !!
def defix (data):
	l = unpack('!H',data[0:2])[0]
	return l,data[2:l+2],data[l+2:]

# README: the name is HORRIBLE, fix this !!
def prefix (data):
	return '%s%s' % (pack('!H',len(data)),data)

class Message (Exception):
	TYPE = None

	MARKER = chr(0xff)*16

	class Type:
		OPEN          = 0x01  #   1
		UPDATE        = 0x02  #   2
		NOTIFICATION  = 0x04  #   4
		KEEPALIVE     = 0x08  #   8
		ROUTE_REFRESH = 0x10  #  16
		LIST          = 0x20  #  32
		HEADER        = 0x40  #  64
		GENERAL       = 0x80  # 128
		#LOCALRIB    = 0x100  # 256

	def __init__ (self):
		if self.TYPE is None:
			self._str = 'UNKNOWN (invalid code)'
			return

		code = ord(self.TYPE)
		result = []

		if code & self.Type.OPEN:
			result.append('OPEN')
		if code & self.Type.UPDATE:
			result.append('UPDATE')
		if code & self.Type.NOTIFICATION:
			result.append('NOTIFICATION')
		if code & self.Type.KEEPALIVE:
			result.append('KEEPALIVE')
		if code & self.Type.ROUTE_REFRESH:
			result.append('ROUTE_REFRESH')
		if code & self.Type.LIST:
			result.append('LIST')
		if code & self.Type.HEADER:
			result.append('HEADER')
		if code & self.Type.GENERAL:
			result.append('GENERAL')

		if result:
			self._str = ' '.join(result)
		else:
			self._str = 'UNKNOWN (%d)' % code

	def _message (self,message = ""):
		message_len = pack('!H',19+len(message))
		return "%s%s%s%s" % (self.MARKER,message_len,self.TYPE,message)

	def __str__ (self):
		return self._str

class Failure (Exception):
	pass



# encoding: utf-8
"""
network.py

Created by Thomas Mangin on 2009-09-06.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
"""

#import os
#import sys
import struct
import time
import socket
#import fcntl
import errno
import select
#import array

from bagpipe.exabgp.utils import hexa,trace
from bagpipe.exabgp.structure.address import AFI
from bagpipe.exabgp.message import Failure

from bagpipe.exabgp.log import Logger,LazyFormat
logger = Logger()

# If the OS tells us we have data on the socket,
# we should never have to wait more than READ_TIMEOUT to be able to read it.
READ_TIMEOUT = 1

errno_block = set((
	errno.EINPROGRESS, errno.EALREADY,
	errno.EAGAIN, errno.EWOULDBLOCK,
	errno.EINTR, errno.EDEADLK,
))

errno_fatal = set((
	errno.EBADF, errno.ECONNRESET,
	errno.ENOTCONN, errno.ESHUTDOWN,
	errno.ECONNABORTED, errno.EPIPE,
))

class Connection (object):
	def __init__ (self,peer,local,md5,ttl):
		self.io = None
		self.last_read = 0
		self.last_write = 0
		self.peer = peer
		self._loop_start = None

		self._buffer = []

		logger.wire("Opening connection to %s" % self.peer)

		if peer.afi != local.afi:
			raise Failure('The local IP and peer IP must be of the same family (both IPv4 or both IPv6)')

		try:
			if peer.afi == AFI.ipv4:
				self.io = socket.socket(socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP)
			if peer.afi == AFI.ipv6:
				self.io = socket.socket(socket.AF_INET6, socket.SOCK_STREAM, socket.IPPROTO_TCP)
			try:
				self.io.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
			except AttributeError:
				pass
			try:
				self.io.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
			except AttributeError:
				pass
			try:
				# diable Nagle's algorithm (no grouping of packets)
				self.io.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
			except AttributeError:
				logger.warning("wire","Could not disable nagle's algorithm for %s" % self.peer)
				pass
			self.io.settimeout(1)
			if peer.afi == AFI.ipv4:
				self.io.bind((local.ip,0))
			if peer.afi == AFI.ipv6:
				self.io.bind((local.ip,0,0,0))
		except socket.error,e:
			self.close()
			raise Failure('Could not bind to local ip %s - %s' % (local.ip,str(e)))

		if md5:
			try:
				TCP_MD5SIG = 14
				TCP_MD5SIG_MAXKEYLEN = 80
				SS_PADSIZE = 120

				n_addr = socket.inet_aton(peer.ip)
				n_port = socket.htons(179)
				tcp_md5sig = 'HH4s%dx2xH4x%ds' % (SS_PADSIZE, TCP_MD5SIG_MAXKEYLEN)
				md5sig = struct.pack(tcp_md5sig, socket.AF_INET, n_port, n_addr, len(md5), md5)
				self.io.setsockopt(socket.IPPROTO_TCP, TCP_MD5SIG, md5sig)
			except socket.error,e:
				self.close()
				raise Failure('This OS does not support TCP_MD5SIG, you can not use MD5 : %s' % str(e))

		# None (ttl-security unset) or zero (maximum TTL) is the same thing
		if ttl:
			try:
				self.io.setsockopt(socket.IPPROTO_IP,socket.IP_TTL, 20)
			except socket.error,e:
				self.close()
				raise Failure('This OS does not support IP_TTL (ttl-security), you can not use MD5 : %s' % str(e))

		try:
			if peer.afi == AFI.ipv4:
				self.io.connect((peer.ip,179))
			if peer.afi == AFI.ipv6:
				self.io.connect((peer.ip,179,0,0))
			self.io.setblocking(0)
		except socket.error, e:
			self.close()
			raise Failure('Could not connect to peer: %s' % str(e))

		try:
			try:
				# Linux / Windows
				self.message_size = self.io.getsockopt(socket.SOL_SOCKET, socket.SO_MAX_MSG_SIZE)
			except AttributeError:
				# BSD
				self.message_size = self.io.getsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF)
		except socket.error, e:
			self.message_size = None

	def pending (self,reset=False):
		if reset:
			self._loop_start = None
		else:
			if not self._loop_start:
				self._loop_start = time.time()
			else:
				if self._loop_start + READ_TIMEOUT < time.time():
					raise Failure('Waited for data on a socket for more than %d second(s)' % READ_TIMEOUT)
		try:
			r,_,_ = select.select([self.io,],[],[],0)
		except select.error,e:
			errno,message = e.args
			if errno in errno_block:
				return False
			raise
		if r: return True
		return False

	def ready (self):
		try:
			_,w,_ = select.select([],[self.io,],[],0)
		except select.error,e:
			errno,message = e.args
			if errno in errno_block:
				return False
			raise
		if not w: return False
		return True

	def read (self,number):
		if not self.io:
			raise Failure('Trying to read on a closed TCP connection')
		if number == 0: return ''
		try:
			r = self.io.recv(number)
			self.last_read = time.time()
			logger.wire(LazyFormat("%15s RECV " % self.peer,hexa,r))
			return r
		except socket.timeout,e:
			self.close()
			raise Failure('Timeout while reading data from the network:  %s ' % str(e))
		except socket.error,e:
			self.close()
			raise Failure('Problem while reading data from the network:  %s ' % str(e))

	def write (self,data):
		if not self.io:
			# We alrady returned a Failure
			# It must be a write attempted during the closing of the peering session
			# Make sure it does not hold the cleanup.
			return True
		if not self.ready():
			return False
		try:
			logger.wire(LazyFormat("%15s SENT " % self.peer,hexa,data))
			self.io.sendall(data)
			self.last_write = time.time()
			return True
		except socket.error, e:
			# Must never happen as we are performing a select before the write
			#failure = getattr(e,'errno',None)
			#if failure in errno_block:
			#	return False
			self.close()
			logger.wire("%15s %s" % (self.peer,trace()))
			raise Failure('Problem while writing data to the network: %s' % str(e))

	def close (self):
		try:
			logger.wire("Closing connection to %s" % self.peer)
			if self.io:
				self.io.close()
		except socket.error:
			pass




# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging

import random

from threading import Lock

from bagpipe.bgp.common import utils
from bagpipe.bgp.common.looking_glass import LookingGlass

log = logging.getLogger(__name__)


class LabelAllocator(LookingGlass):
    # Warning: does not reuse labels, will break if more than 2**20-16 labels
    # are requested along the life-cycle
    # FIXME: add code so that all the labels that were cleaned'up will *not*
    # be reused in a short time

    def __init__(self):
        self.currentLabel = random.randint(
            100, 200)  # Labels below 16 are reserved
        # using a random start value will illustrate during demos and tests
        # that the label for a VRF does not
        # need be the same on all compute nodes
        self.labels = dict()

        self.lock = Lock()

    @utils.synchronized
    def getNewLabel(self, description):

        if (self.currentLabel == 2 ** 20):
            # Looking forward to the day will hit this one:
            log.error("All the 2^20 possible labels have been used at least "
                      "once, and this piece of code doesn't know how to reuse "
                      "them")
            raise Exception("Out of labels")

        label = self.currentLabel
        self.currentLabel += 1
        self.labels[label] = description

        log.debug("Allocated label %d for '%s'", label, description)
        return label

    @utils.synchronized
    def release(self, label):
        if label in self.labels:
            log.debug("released label %d ('%s')", label, self.labels[label])
            del self.labels[label]
        else:
            log.warn("asked to release a non registered label: %d", label)

    def getLookingGlassLocalInfo(self, prefix):
        return self.labels



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from threading import Lock

import re
import logging

from bagpipe.bgp.vpn.ipvpn import VRF
from bagpipe.bgp.vpn.evpn import EVI

import bagpipe.bgp.common.exceptions as exc

from bagpipe.bgp.common.looking_glass import LookingGlass, LGMap
from bagpipe.bgp.common import utils
from bagpipe.bgp.common import logDecorator
from bagpipe.bgp.common.run_command import runCommand

from bagpipe.bgp.vpn.label_allocator import LabelAllocator

from bagpipe.exabgp.message.update.attribute.communities import RouteTarget


log = logging.getLogger(__name__)


def convertRouteTargets(orig_list):
    assert(isinstance(orig_list, list))
    list_ = []
    for rt in orig_list:
        if rt == '':
            continue
        try:
            asn, nn = rt.split(':')
            list_.append(RouteTarget(int(asn), None, int(nn)))
        except Exception:
            raise Exception("Malformed route target: '%s'" % rt)
    return list_


class VPNManager(LookingGlass):

    """
    Creates, and keeps track of, VPN instances (VRFs and EVIs) and passes
    plug/unplug calls to the right VPN instance.
    """

    type2class = {"ipvpn": VRF,
                  "evpn": EVI
                  }

    @logDecorator.log
    def __init__(self, bgpManager, dataplaneDrivers):
        '''
        dataplaneDrivers is a dict from vpn type to each dataplane driver,
        e.g. { "ipvpn": driverA, "evpn": driverB }
        '''

        self.bgpManager = bgpManager

        self.dataplaneDrivers = dataplaneDrivers

        # Init VPN instance identifiers
        self.instanceId = 1

        # VPN instance dict
        self.vpnInstances = {}

        logging.debug("Creating label allocator")
        self.labelAllocator = LabelAllocator()

        # dict containing info how an ipvpn is plugged
        # from an evpn  (keys: ipvpn instances)
        self._evpn_ipvpn_ifs = {}

        self.lock = Lock()

    def _formatIpAddressPrefix(self, ipAddress):
        if re.match(r'([12]?\d?\d\.){3}[12]?\d?\d\/[123]?\d', ipAddress):
            address = ipAddress
        elif re.match(r'([12]?\d?\d\.){3}[12]?\d?\d', ipAddress):
            address = ipAddress + "/32"
        else:
            raise exc.MalformedIPAddress

        return address

    @utils.synchronized
    def getInstanceId(self):
        iid = self.instanceId
        self.instanceId += 1
        return iid

    @logDecorator.logInfo
    def _attach_evpn2ipvpn(self, localPort, ipvpnInstance):
        """ Assuming localPort indicates no real interface but only
        an EVPN, this method will create a pair of twin interfaces, one
        to plug in the EVPN, the other to plug in the IPVPN.

        The localPort dict will be modified so that the 'linuxif' indicates
        the name of the interface to plug in the IPVPN.

        The EVPN instance will be notified so that it forwards traffic
        destinated to the gateway on the interface toward the IPVPN.
        """
        assert('evpn' in localPort)

        if 'id' not in localPort['evpn']:
            raise Exception("Missing parameter 'id' :an external EVPN "
                            "instance id must be specified for an EVPN "
                            "attachment")

        try:
            evpn = self.vpnInstances[localPort['evpn']['id']]
        except:
            raise Exception("The specified evpn instance does not exist (%s)"
                            % localPort['evpn'])

        if (evpn.type != "evpn"):
            raise Exception("The specified instance to plug is not an epnv"
                            "instance (is %s instead)" % evpn.type)

        if ipvpnInstance in self._evpn_ipvpn_ifs:
            (evpn_if, ipvpn_if, evpn, managed) = \
                self._evpn_ipvpn_ifs[ipvpnInstance]

            if not (localPort['evpn']['id'] == evpn.externalInstanceId):
                raise Exception('Trying to plug into an IPVPN a new E-VPN '
                                'while one is already plugged in')
            else:
                # do nothing
                log.warning('Trying to plug an E-VPN into an IPVPN, but it was'
                            'already done')
                localPort['linuxif'] = ipvpn_if
                return

        #  detect if this evpn is already plugged into an IPVPN
        if evpn.hasGatewayPort():
            raise Exception("Trying to plug E-VPN into an IPVPN, but this EVPN"
                            " is already plugged into an IPVPN")

        if ('linuxif' in localPort and localPort['linuxif']):
            raise Exception("Cannot specify an attachment with both a linuxif "
                            "and an evpn")

        if 'ovs_port_name' in localPort['evpn']:
            try:
                assert(localPort['ovs']['plugged'])
                assert(localPort['ovs']['port_name'] or
                       localPort['ovs']['port_number'])
            except:
                raise Exception("Using ovs_port_name in EVPN/IPVPN attachment"
                                " requires specifying the corresponding OVS"
                                " port, which must also be pre-plugged")

            evpn_if = localPort['evpn']['ovs_port_name']

            # we assume in this case that the E-VPN interface is already
            # plugged into the E-VPN bridge
            managed = False
        else:
            evpn_if = "evpn%d-ipvpn%d" % (
                evpn.instanceId, ipvpnInstance.instanceId)
            ipvpn_if = "ipvpn%d-evpn%d" % (
                ipvpnInstance.instanceId, evpn.instanceId)

            # FIXME: do it only if not existing already...
            log.info("Creating veth pair %s %s ", evpn_if, ipvpn_if)

            # delete the interfaces if they exist already
            runCommand(log, "ip link delete %s" %
                       evpn_if, acceptableReturnCodes=[0, 1])
            runCommand(log, "ip link delete %s" %
                       ipvpn_if, acceptableReturnCodes=[0, 1])

            runCommand(log, "ip link add %s type veth peer name %s" %
                       (evpn_if, ipvpn_if))

            runCommand(log, "ip link set %s up" % evpn_if)
            runCommand(log, "ip link set %s up" % ipvpn_if)
            managed = True

        localPort['linuxif'] = ipvpn_if

        evpn.setGatewayPort(evpn_if, ipvpnInstance)

        self._evpn_ipvpn_ifs[ipvpnInstance] = (
            evpn_if, ipvpn_if, evpn, managed)

    @logDecorator.logInfo
    def _detach_evpn2ipvpn(self, ipvpn):
        """
        Symmetric to _attach_evpn2ipvpn
        """
        (evpn_if, ipvpn_if, evpnInstance,
         managed) = self._evpn_ipvpn_ifs[ipvpn]

        if not ipvpn.hasEnpoint(ipvpn_if):
            # TODO: check that this evpn instance is still up and running
            evpnInstance.gatewayPortDown(evpn_if)

            # cleanup veth pair
            if managed:
                runCommand(log, "ip link delete %s" % evpn_if)

            del self._evpn_ipvpn_ifs[ipvpn]

    def _cleanup_evpn2ipvpn(self, ipvpn):
        (_, ipvpn_if, _, managed) = self._evpn_ipvpn_ifs[ipvpn]

        # cleanup veth pair
        if managed:
            runCommand(log, "ovs-vsctl del-port %s" % ipvpn_if)
            runCommand(log, "ip link delete %s" % ipvpn_if)

    @logDecorator.logInfo
    def plugVifToVPN(self, externalInstanceId, instanceType, importRTs,
                     exportRTs, macAddress, ipAddress, gatewayIP,
                     localPort, linuxbr, advertiseSubnet, readvertise,
                     fallback):

        # Verify and format IP address with prefix if necessary
        try:
            ipAddressPrefix = self._formatIpAddressPrefix(ipAddress)
        except exc.MalformedIPAddress:
            raise

        # Convert route target string to RouteTarget dictionary
        importRTs = convertRouteTargets(importRTs)
        exportRTs = convertRouteTargets(exportRTs)

        if readvertise:
            try:
                readvertise = {k: convertRouteTargets(readvertise[k])
                               for k in ['from_rt', 'to_rt']}
            except KeyError as e:
                raise Exception("Wrong 'readvertise' parameters: %s" % e)

        # retrieve network mask
        mask = int(ipAddressPrefix.split('/')[1])

        # Retrieve VPN instance or create new one if does not exist
        try:
            vpnInstance = self.vpnInstances[externalInstanceId]
            if (vpnInstance.type != instanceType):
                raise Exception("Trying to plug port on an existing instance "
                                "of a different type (existing: %s, asked: %s)"
                                % (vpnInstance.type, instanceType))
        except KeyError:
            instanceId = self.getInstanceId()
            log.info("Create and start new VPN instance %d for external "
                     "network instance identifier %s", instanceId,
                     externalInstanceId)
            try:
                vpnInstanceFactory = VPNManager.type2class[instanceType]
            except KeyError:
                log.error("Unsupported instanceType for VPNInstance: %s",
                          instanceType)
                raise Exception("Unsupported instance type: %s" % instanceType)

            try:
                dataplaneDriver = self.dataplaneDrivers[instanceType]
            except KeyError:
                log.error("No dataplane driver for VPN type %s",
                          instanceType)
                raise Exception("No dataplane driver for VPN type %s" %
                                instanceType)

            if instanceType == "evpn" and linuxbr:
                kwargs = {'linuxbr': linuxbr}
            else:
                kwargs = {}

            vpnInstance = vpnInstanceFactory(
                self.bgpManager, self.labelAllocator, dataplaneDriver,
                externalInstanceId, instanceId, importRTs, exportRTs,
                gatewayIP, mask, readvertise, fallback, **kwargs)

            # Update VPN instance list
            self.vpnInstances[externalInstanceId] = vpnInstance

            vpnInstance.start()

        # Check if new route target import/export must be updated
        if not ((set(vpnInstance.importRTs) == set(importRTs)) and
                (set(vpnInstance.exportRTs) == set(exportRTs))):
            vpnInstance.updateRouteTargets(importRTs, exportRTs)

        vpnInstance.update_fallback(fallback)

        if instanceType == "ipvpn" and 'evpn' in localPort:
            # special processing for the case where what we plug into
            # the ipvpn is not an existing interface but an interface
            # to create, connected to an existing evpn instance
            self._attach_evpn2ipvpn(localPort, vpnInstance)

        # Plug VIF to VPN instance
        vpnInstance.vifPlugged(macAddress, ipAddressPrefix, localPort,
                               advertiseSubnet)

    @logDecorator.logInfo
    def unplugVifFromVPN(self, externalInstanceId, macAddress, ipAddress,
                         localPort, readvertise):

        # Verify and format IP address with prefix if necessary
        try:
            ipAddressPrefix = self._formatIpAddressPrefix(ipAddress)
        except exc.MalformedIPAddress:
            raise

        # Retrieve VPN instance or raise exception if does not exist
        try:
            vpnInstance = self.vpnInstances[externalInstanceId]
        except KeyError:
            log.error("Try to unplug VIF from non existing VPN instance %s",
                      externalInstanceId)
            raise exc.VPNNotFound(externalInstanceId)

        # Unplug VIF from VPN instance
        vpnInstance.vifUnplugged(macAddress, ipAddressPrefix, readvertise)

        if vpnInstance.type == "ipvpn" and 'evpn' in localPort:
            self._detach_evpn2ipvpn(vpnInstance)

        if vpnInstance.stopIfEmpty():
            del self.vpnInstances[externalInstanceId]

    @logDecorator.logInfo
    def stop(self):
        for vpnInstance in self.vpnInstances.itervalues():
            vpnInstance.stop()
            # Cleanup veth pair
            if (vpnInstance.type == "ipvpn" and
                    self._evpn_ipvpn_ifs.get(vpnInstance)):
                self._cleanup_evpn2ipvpn(vpnInstance)
        for vpnInstance in self.vpnInstances.itervalues():
            vpnInstance.join()

    # Looking Glass hooks ####

    def getLGMap(self):
        class DataplaneLGHook(LookingGlass):

            def __init__(self, vpnManager):
                self.vpnManager = vpnManager

            def getLGMap(self):
                return {
                    "drivers": (LGMap.COLLECTION, (
                        self.vpnManager.getLGDataplanesList,
                        self.vpnManager.getLGDataplaneFromPathItem)),
                    "ids": (LGMap.DELEGATE, self.vpnManager.labelAllocator)
                }
        dataplaneHook = DataplaneLGHook(self)
        return {
            "instances": (LGMap.COLLECTION, (self.getLGVPNList,
                                             self.getLGVPNFromPathItem)),
            "dataplane": (LGMap.DELEGATE, dataplaneHook)
        }

    def getLGVPNList(self):
        return [{"id": id,
                 "name": instance.name}
                for (id, instance) in self.vpnInstances.iteritems()]

    def getLGVPNFromPathItem(self, pathItem):
        return self.vpnInstances[pathItem]

    def getVPNInstancesCount(self):
        return len(self.vpnInstances)

    # LookingGLass ########

    def getLGDataplanesList(self):
        return [{"id": i} for i in self.dataplaneDrivers.iterkeys()]

    def getLGDataplaneFromPathItem(self, pathItem):
        return self.dataplaneDrivers[pathItem]



# encoding: utf-8
"""
exabgp.py

Created by Thomas Mangin on 2009-08-30.
Copyright (c) 2009 Exa Networks. All rights reserved.
"""

import os
import sys
import time
import signal
import select

from bagpipe.exabgp.network.peer import Peer
from bagpipe.exabgp.version import version

from bagpipe.exabgp.daemon import Daemon
from bagpipe.exabgp.processes import Processes
from bagpipe.exabgp.configuration import Configuration
from bagpipe.exabgp.network.connection import errno_block

from bagpipe.exabgp.processes import ProcessError

from bagpipe.exabgp.log import Logger
logger = Logger()


class Supervisor (object):
	# [hex(ord(c)) for c in os.popen('clear').read()]
	clear = ''.join([chr(int(c,16)) for c in ['0x1b', '0x5b', '0x48', '0x1b', '0x5b', '0x32', '0x4a']])

	def __init__ (self,configuration):
		self.daemon = Daemon(self)
		self.processes = Processes(self)
		self.configuration = Configuration(configuration)

		self.watchdogs = {}
		self._peers = {}
		self._shutdown = False
		self._reload = False
		self._restart = False
		self._route_update = False
		self._commands = {}
		self._saved_pid = False
		self.reload()

		signal.signal(signal.SIGTERM, self.sigterm)
		signal.signal(signal.SIGHUP, self.sighup)
		signal.signal(signal.SIGALRM, self.sigalrm)

	def sigterm (self,signum, frame):
		logger.supervisor("SIG TERM received")
		self._shutdown = True

	def sighup (self,signum, frame):
		logger.supervisor("SIG HUP received")
		self._reload = True

	def sigalrm (self,signum, frame):
		logger.supervisor("SIG ALRM received")
		self._restart = True

	def run (self,supervisor_speed=0.5):
		if self.daemon.drop_privileges():
			logger.supervisor("Could not drop privileges to '%s' refusing to run as root" % self.daemon.user)
			logger.supervisor("Set the environmemnt value USER to change the unprivileged user")
			return
		self.daemon.daemonise()
		self.daemon.savepid()

		# did we complete the run of updates caused by the last SIGHUP ?
		reload_completed = True

		while True:
			try:
				while self._peers:
					start = time.time()

					self.handle_commands(self.processes.received())

					if self._shutdown:
						self._shutdown = False
						self.shutdown()
					elif self._reload and reload_completed:
						self._reload = False
						self.reload()
					elif self._restart:
						self._restart = False
						self.restart()
					elif self._route_update:
						self._route_update = False
						self.route_update()
					elif self._commands:
						self.commands(self._commands)
						self._commands = {}

					reload_completed = True
					# Handle all connection
					peers = self._peers.keys()
					ios = []
					while peers:
						for key in peers[:]:
							peer = self._peers[key]
							# there was no routes to send for this peer, we performed keepalive checks
							if peer.run() is not True:
								# no need to come back to it before a a full cycle
								if peer.bgp and peer.bgp.connection:
									ios.append(peer.bgp.connection.io)
								peers.remove(key)
							# send the route we parsed (if we parsed any to our child processes)
							# This is a generator and can only be run once
							try:
								for route in peer.received_routes():
									# This is a generator which content does only change at config reload
									for name in self.processes.receive_routes():
										# using str(key) as we should not manipulate it and assume its format
										self.processes.write(name,'neighbor %s %s\n' % (str(key),route))
							except ProcessError:
								# Can not find any better error code that 6,0 !
								raise Notify(6,0,'ExaBGP Internal error, sorry.')
						# otherwise process as many routes as we can within a second for the remaining peers
						duration = time.time() - start
						# RFC state that we MUST not more than one KEEPALIVE / sec
						# And doing less could cause the session to drop
						if duration >= 1.0:
							reload_completed = False
							ios=[]
							break
					duration = time.time() - start
					if ios:
						try:
							read,_,_ = select.select(ios,[],[],max(supervisor_speed-duration,0))
						except select.error,e:
							errno,message = e.args
							if not errno in errno_block:
								raise
					else:
						if duration < supervisor_speed:
							time.sleep(max(supervisor_speed-duration,0))
				self.processes.terminate()
				self.daemon.removepid()
				break
			except KeyboardInterrupt:
				logger.supervisor("^C received")
				self._shutdown = True
			except IOError:
				logger.supervisor("I/O Error received, most likely ^C during IO")
				self._shutdown = True
			except ProcessError:
				logger.supervisor("Problem when sending message(s) to helper program, stopping")
				self._shutdown = True
#				from leak import objgraph
#				print objgraph.show_most_common_types(limit=20)
#				import random
#				obj = objgraph.by_type('ReceivedRoute')[random.randint(0,2000)]
#				objgraph.show_backrefs([obj], max_depth=10)

	def shutdown (self):
		"""terminate all the current BGP connections"""
		logger.info("Performing shutdown","supervisor")
		for key in self._peers.keys():
			self._peers[key].stop()

	def reload (self):
		"""reload the configuration and send to the peer the route which changed"""
		logger.info("Performing reload of exabgp %s" % version,"configuration")

		reloaded = self.configuration.reload()
		if not reloaded:
			logger.info("Problem with the configuration file, no change done","configuration")
			logger.info(self.configuration.error,"configuration")
			return

		for key in self._peers.keys():
			if key not in self.configuration.neighbor.keys():
				neighbor = self.configuration.neighbor[key]
				logger.supervisor("Removing Peer %s" % neighbor.name())
				self._peers[key].stop()

		for key in self.configuration.neighbor.keys():
			neighbor = self.configuration.neighbor[key]
			# new peer
			if key not in self._peers.keys():
				logger.supervisor("New Peer %s" % neighbor.name())
				peer = Peer(neighbor,self)
				self._peers[key] = peer
			else:
				# check if the neighbor definition are the same (BUT NOT THE ROUTES)
				if self._peers[key].neighbor != neighbor:
					logger.supervisor("Peer definition change, restarting %s" % str(key))
					self._peers[key].restart(neighbor)
				# set the new neighbor with the new routes
				else:
					logger.supervisor("Updating routes for peer %s" % str(key))
					self._peers[key].reload(neighbor.every_routes())
		logger.info("Loaded new configuration successfully",'configuration')
		self.processes.start()

	def handle_commands (self,commands):
		for service in commands:
			for command in commands[service]:
				# watchdog
				if command.startswith('announce watchdog') or command.startswith('withdraw watchdog'):
					parts = command.split(' ')
					try:
						name = parts[2]
					except IndexError:
						name = service
					self.watchdogs[name] = parts[0]
					self._route_update = True

				# route announcement / withdrawal
				elif command.startswith('announce route'):
					route = self.configuration.parse_single_route(command)
					if not route:
						logger.supervisor("Command could not parse route in : %s" % command)
					else:
						self.configuration.add_route_all_peers(route)
						self._route_update = True
				elif command.startswith('withdraw route'):
					route = self.configuration.parse_single_route(command)
					if not route:
						logger.supervisor("Command could not parse route in : %s" % command)
					else:
						if self.configuration.remove_route_all_peers(route):
							logger.supervisor("Command success, route found and removed : %s" % route)
							self._route_update = True
						else:
							logger.supervisor("Command failure, route not found : %s" % route)


				# flow announcement / withdrawal
				elif command.startswith('announce flow'):
					flow = self.configuration.parse_single_flow(command)
					if not flow:
						logger.supervisor("Command could not parse flow in : %s" % command)
					else:
						self.configuration.add_route_all_peers(flow)
						self._route_update = True
				elif command.startswith('withdraw flow'):
					flow = self.configuration.parse_single_flow(command)
					if not flow:
						logger.supervisor("Command could not parse flow in : %s" % command)
					else:
						if self.configuration.remove_route_all_peers(flow):
							logger.supervisor("Command success, flow found and removed : %s" % flow)
							self._route_update = True
						else:
							logger.supervisor("Command failure, flow not found : %s" % flow)

				# commands
				elif command in ['reload','restart','shutdown','version']:
					self._commands.setdefault(service,[]).append(command)

				# unknown
				else:
					logger.supervisor("Command from process not understood : %s" % command)

	def commands (self,commands):
		def _answer (service,string):
			self.processes.write(service,string)
			logger.supervisor('Responding to %s : %s' % (service,string))

		for service in commands:
			for command in commands[service]:
				if command == 'shutdown':
					self._shutdown = True
					_answer(service,'shutdown in progress')
					continue
				if command == 'reload':
					self._reload = True
					_answer(service,'reload in progress')
					continue
				if command == 'restart':
					self._restart = True
					_answer(service,'restart in progress')
					continue
				if command == 'version':
					_answer(service,'exabgp %s' % version)
					continue

	def route_update (self):
		"""the process ran and we need to figure what routes to changes"""
		logger.supervisor("Performing dynamic route update")

		for key in self.configuration.neighbor.keys():
			neighbor = self.configuration.neighbor[key]
			neighbor.watchdog(self.watchdogs)
			self._peers[key].reload(neighbor.every_routes())
		logger.supervisor("Updated peers dynamic routes successfully")

	def restart (self):
		"""kill the BGP session and restart it"""
		logger.info("Performing restart of exabgp %s" % version,"supervisor")
		self.configuration.reload()

		for key in self._peers.keys():
			if key not in self.configuration.neighbor.keys():
				neighbor = self.configuration.neighbor[key]
				logger.supervisor("Removing Peer %s" % neighbor.name())
				self._peers[key].stop()
			else:
				self._peers[key].restart()
		self.processes.start()

	def unschedule (self,peer):
		key = peer.neighbor.name()
		if key in self._peers:
			del self._peers[key]


def version_warning ():
	sys.stdout.write('\n')
	sys.stdout.write('************ WARNING *** WARNING *** WARNING *** WARNING *********\n')
	sys.stdout.write('* This program SHOULD work with your python version (2.4).       *\n')
	sys.stdout.write('* No tests have been performed. Consider python 2.4 unsupported  *\n')
	sys.stdout.write('* Please consider upgrading to the latest 2.x stable realease.   *\n')
	sys.stdout.write('************ WARNING *** WARNING *** WARNING *** WARNING *********\n')
	sys.stdout.write('\n')

def help ():
	sys.stdout.write('\n')
	sys.stdout.write('*******************************************************************************\n')
	sys.stdout.write('set the following environment values to gather information and report bugs\n')
	sys.stdout.write('\n')
	sys.stdout.write('DEBUG_ALL : debug everything\n')
	sys.stdout.write('DEBUG_CONFIGURATION : verbose configuration parsing\n')
	sys.stdout.write('DEBUG_SUPERVISOR : signal received, configuration reload (default: yes))\n')
	sys.stdout.write('DEBUG_DAEMON : pid change, forking, ... (default: yes))\n')
	sys.stdout.write('DEBUG_PROCESSES : handling of forked processes (default: yes))\n')
	sys.stdout.write('DEBUG_WIRE : the packet sent and received\n')
	sys.stdout.write('DEBUG_RIB : change in route announcement in config reload\n')
	sys.stdout.write('DEBUG_MESSAGE : changes in route announcement in config reload (default: yes)\n')
	sys.stdout.write('DEBUG_TIMERS : tracking keepalives\n')
	sys.stdout.write('DEBUG_ROUTES : print parsed routes\n')
	sys.stdout.write('\n')
	sys.stdout.write('PROFILE : (1,true,on,yes,enable) profiling info on exist\n')
	sys.stdout.write('          use a filename to dump the outpout in a file\n')
	sys.stdout.write('          IMPORTANT : exabpg will not overwrite existing files\n')
	sys.stdout.write('\n')
	sys.stdout.write('PDB : on program fault, start pdb the python interactive debugger\n')
	sys.stdout.write('\n')
	sys.stdout.write('USER : the user the program should try to use if run by root (default: nobody)\n')
	sys.stdout.write('PID : the file in which the pid of the program should be stored\n')
	sys.stdout.write('SYSLOG: no value for local syslog, a file name (which will auto-rotate) or host:<host> for remote syslog\n')
	sys.stdout.write('DAEMONIZE: detach and send the program in the background\n')
	sys.stdout.write('MINIMAL_MP: when negociating multiprotocol, try to announce as few AFI/SAFI pair as possible\n')
	sys.stdout.write('\n')
	sys.stdout.write('For example :\n')
	sys.stdout.write('> env PDB=1 PROFILE=~/profile.log DEBUG_SUPERVISOR=0 DEBUG_WIRE=1 \\\n')
	sys.stdout.write('     USER=wheel SYSLOG=host:127.0.0.1 DAEMONIZE= PID=/var/run/exabpg.pid \\\n')
	sys.stdout.write('     ./bin/exabgp ./etc/bgp/configuration.txt\n')
	sys.stdout.write('*******************************************************************************\n')
	sys.stdout.write('\n')
	sys.stdout.write('usage:\n exabgp <configuration file>\n')

def main ():
	main = int(sys.version[0])
	secondary = int(sys.version[2])

	if main != 2 or secondary < 4:
		sys.exit('This program can not work (is not tested) with your python version (< 2.4 or >= 3.0)')

	if main == 2 and secondary == 4:
		version_warning()

	if len(sys.argv) < 2:
		help()
		sys.exit(0)

	for arg in sys.argv[1:]:
		if arg in ['--',]:
			break
		if arg in ['-h','--help']:
			help()
			sys.exit(0)

	Supervisor(sys.argv[1]).run()
	sys.exit(0)

if __name__ == '__main__':
	profiled = os.environ.get('PROFILE',0)
	if profiled == 0:
		main()
	else:
		try:
			import cProfile as profile
		except:
			import profile
		if profiled.lower() in ['1','true','yes','on','enable']:
			profile.run('main()')
		else:
			notice = ''
			if os.path.isdir(profiled):
				notice = 'profile can not use this filename as outpout, it is not a directory (%s)' % profiled
			if os.path.exists(profiled):
				notice = 'profile can not use this filename as outpout, it already exists (%s)' % profiled

			if not notice:
				logger.supervisor('profiling ....')
				profile.run('main()',filename=profiled)
			else:
				logger.supervisor("-"*len(notice))
				logger.supervisor(notice)
				logger.supervisor("-"*len(notice))
				main()



# encoding: utf-8
"""
mprnlri.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
"""

from bagpipe.exabgp.structure.address import Address
from bagpipe.exabgp.message.update.attribute import AttributeID,Flag,Attribute

# =================================================================== MP NLRI (14)

class MPURNLRI (Attribute):
	FLAG = Flag.OPTIONAL
	ID = AttributeID.MP_UNREACH_NLRI
	MULTIPLE = True

	def __init__ (self,routes):
		self.routes = routes

	def pack (self):
		return self._attribute(self.routes[0].nlri.afi.pack() + self.routes[0].nlri.safi.pack() + ''.join([route.nlri.pack() for route in self.routes]))

	def __len__ (self):
		return len(self.pack())

	def __str__ (self):
		return "MP_UNREACH_NLRI Family %s %d NLRI(s)" % (Address.__str__(self.routes[0]),len(self.routes))

	def __repr__ (self):
		return str(self)



# encoding: utf-8
"""
set.py

Created by Thomas Mangin on 2010-01-16.
Copyright (c) 2010-2012 Exa Networks. All rights reserved.
Modified by Orange - 2014
"""

import collections

from bagpipe.exabgp.structure.address import AFI,SAFI
from bagpipe.exabgp.structure.asn import AS_TRANS
from bagpipe.exabgp.message.update.attribute import AttributeID

from bagpipe.exabgp.message.update.attribute.origin      import Origin
from bagpipe.exabgp.message.update.attribute.aspath      import ASPath,AS4Path
from bagpipe.exabgp.message.update.attribute.localpref   import LocalPreference

# =================================================================== Attributes

class MultiAttributes (list):
	def __init__ (self,attribute):
		list.__init__(self)
		self.ID = attribute.ID
		self.FLAG = attribute.FLAG
		self.MULTIPLE = True
		self.append(attribute)

	def pack (self):
		r = []
		for attribute in self:
			r.append(attribute.pack())
		return ''.join(r)

	def __len__ (self):
		return len(self.pack())

	def __str__ (self):
		return 'MultiAttibutes(%s)' % ' '.join(str(_) for _ in self)

	def __repr__ (self):
		return str(self)

class Attributes (dict):
	autocomplete = True

	def __init__ (self):
		self._str = ''

	def has (self,k):
		return self.has_key(k)

	def add (self,attribute):
		self._str = ''
		if self.has(attribute.ID):
			if attribute.MULTIPLE:
				self[attribute.ID].append(attribute)
				return True
			return False
		else:
			if attribute.MULTIPLE:
				self[attribute.ID] = MultiAttributes(attribute)
			else:
				self[attribute.ID] = attribute
			return True

	def remove (self,attrid):
		self.pop(attrid)

	def _as_path (self,asn4,asp):
		message = ''
		# if the peer does not understand ASN4, we need to build a transitive AS4_PATH
		if not asn4:
			has_asn4 = False
			aspath = ASPath(False,asp.asptype)
			as4path = AS4Path(asp.asptype)
			for segment in asp.aspsegment:
				if segment.asn4():
					has_asn4 = True
					aspath.add(AS_TRANS)
					as4path.add(segment)
				else:
					aspath.add(segment)
					as4path.add(segment)
			message += aspath.pack()
			if has_asn4:
				message += as4path.pack()
		else:
			message += ASPath(True,asp.asptype,asp.aspsegment).pack()
		return message

	def bgp_announce (self,asn4,local_asn,peer_asn):
		ibgp = (local_asn == peer_asn)
		# we do not store or send MED
		message = ''

		if AttributeID.ORIGIN in self:
			message += self[AttributeID.ORIGIN].pack()
		elif self.autocomplete:
			message += Origin(Origin.IGP).pack()

		if AttributeID.AS_PATH in self:
			asp = self[AttributeID.AS_PATH]
			message += self._as_path(asn4,asp)
		elif self.autocomplete:
			if ibgp:
				asp = ASPath(asn4,ASPath.AS_SEQUENCE,[])
			else:
				asp = ASPath(asn4,ASPath.AS_SEQUENCE,[local_asn])
			message += self._as_path(asn4,asp)
		else:
			raise RuntimeError('Generated routes must always have an AS_PATH ')

		if AttributeID.NEXT_HOP in self:
			afi = self[AttributeID.NEXT_HOP].next_hop.afi
			safi = self[AttributeID.NEXT_HOP].next_hop.safi
			if afi == AFI.ipv4 and safi in [SAFI.unicast, SAFI.multicast]:
				message += self[AttributeID.NEXT_HOP].pack()

		if ibgp:
			if AttributeID.LOCAL_PREF in self:
				message += self[AttributeID.LOCAL_PREF].pack()
			else:
				message += LocalPreference(100).pack()

		if AttributeID.MED in self:
			if local_asn != peer_asn:
				message += self[AttributeID.MED].pack()

		if AttributeID.PMSI_TUNNEL in self:
			message += self[AttributeID.PMSI_TUNNEL].pack()

		for attribute in [AttributeID.COMMUNITY,AttributeID.EXTENDED_COMMUNITY]:
			if attribute in self:
				message += self[attribute].pack()

		return message

	def __str__ (self):
		if self._str:
			return self._str

		next_hop = ''
		if self.has(AttributeID.NEXT_HOP):
			next_hop = ' next-hop %s' % str(self[AttributeID.NEXT_HOP]).lower()

		origin = ''
		if self.has(AttributeID.ORIGIN):
			origin = ' origin %s' % str(self[AttributeID.ORIGIN]).lower()

		aspath = ''
		if self.has(AttributeID.AS_PATH):
			aspath = ' %s' % str(self[AttributeID.AS_PATH]).lower().replace('_','-')

		local_pref = ''
		if self.has(AttributeID.LOCAL_PREF):
			l = self[AttributeID.LOCAL_PREF]
			local_pref = ' local_preference %s' % l

		med = ''
		if self.has(AttributeID.MED):
			m = self[AttributeID.MED]
			med = ' med %s' % m

		communities = ''
		if self.has(AttributeID.COMMUNITY):
			communities = ' community %s' % str(self[AttributeID.COMMUNITY])

		pmsi = ""
		if self.has(AttributeID.PMSI_TUNNEL):
			pmsi = ' pmsi %s' % str(self[AttributeID.PMSI_TUNNEL])

		mpr = ''
		ecommunities = ''
		if self.has(AttributeID.EXTENDED_COMMUNITY):
			ecommunities = ' extended-community %s' % str(self[AttributeID.EXTENDED_COMMUNITY])

		if self.has(AttributeID.MP_REACH_NLRI):
			mpr = ' mp_reach_nlri %s' % str(self[AttributeID.MP_REACH_NLRI])

		self._str = "%s%s%s%s%s%s%s%s%s" % (next_hop,origin,aspath,local_pref,med,communities,ecommunities,pmsi,mpr)
		return self._str


	def __repr__ (self):
		return str(self)
	
	def __hash__(self):
		return hash(repr(self))  #FIXME: not excellent... :-(

	def sameValuesAs(self,other):
		# test that sets of attributes exactly match
		# can't rely on __eq__ for this, because __eq__ relies on Attribute.__eq__ which does not look at attributes values
	
		for key in set(self.iterkeys()).union(set(other.iterkeys())):
			
			if key == AttributeID.MP_REACH_NLRI: continue

			try:
				sval = self[key]
				oval = other[key]
				
				# In the case where the attribute is, for instance, a list
				# we want to compare values independently of the order
				if isinstance(sval, collections.Iterable):
					if not isinstance(oval, collections.Iterable):
						return False
					else:
						# we sort based on string representation since the items do not
						# necessarily implement __cmp__
						sorter=lambda x,y: cmp(repr(x), repr(y))
						sval = sorted(sval,sorter)
						oval = set(oval,sorter)
						
				if sval != oval: 
					return False
			except KeyError:
				return False
			
		return True
	
